# PyZX - Python library for quantum circuit rewriting
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
from collections import defaultdict
from fractions import Fraction

from typing import Callable, Optional, List, Dict, Tuple

from .utils import EdgeType, VertexType, FractionLike
from .utils import toggle_edge, vertex_is_zx, toggle_vertex
from .graph.base import BaseGraph, VT, ET, upair
import pyzx_param.rewrite_rules.rules as rules
import pyzx_param.rewrite_rules.hrules as hrules


def match_X_spiders(
        g: BaseGraph[VT, ET],
        vertexf: Optional[Callable[[VT], bool]] = None
        ) -> List[VT]:
    if vertexf is not None: candidates = set([v for v in g.vertices() if vertexf(v)])
    else: candidates = g.vertex_set()
    types = g.types()

    return [v for v in candidates if types[v] == VertexType.X]


def match_Z_spiders(
        g: BaseGraph[VT, ET],
        vertexf: Optional[Callable[[VT], bool]] = None
        ) -> List[VT]:
    if vertexf is not None: candidates = set([v for v in g.vertices() if vertexf(v)])
    else: candidates = g.vertex_set()
    types = g.types()

    return [v for v in candidates if types[v] == VertexType.Z]


def color_change(g: BaseGraph[VT,ET], matches: List[VT]) -> rules.RewriteOutputType[VT,ET]:
    for v in matches:
        g.set_type(v, toggle_vertex(g.type(v)))
        for e in g.incident_edges(v):
            g.set_edge_type(e, toggle_edge(g.edge_type(e)))
    return ({}, [],[],False)


def pauli_matcher(
        g: BaseGraph[VT,ET], 
        vertexf: Optional[Callable[[VT],bool]] = None
        ) -> List[Tuple[VT,VT]]:
    if vertexf is not None: candidates = set([v for v in g.vertices() if vertexf(v)])
    else: candidates = g.vertex_set()
    phases = g.phases()
    types = g.types()
    m: List[Tuple[VT,VT]] = []
    paulis = {v for v in candidates
                if phases[v] == 1 and vertex_is_zx(types[v])}
    if not paulis: return m
    while len(candidates) > 0:
        v = candidates.pop()
        if v in paulis and g.vertex_degree(v) == 2: continue
        for w in g.neighbors(v):
            if w in paulis: break
        else:
            continue
        et = g.edge_type(g.edge(v,w))
        if ((types[v] == types[w] and et == EdgeType.HADAMARD) or
            (vertex_is_zx(types[v]) and types[v] != types[w] and et == EdgeType.SIMPLE) or
            (types[v] == VertexType.H_BOX and phases[v] == 1 and (
                (et == EdgeType.SIMPLE and types[w] == VertexType.X) or
                (et == EdgeType.HADAMARD and types[w] == VertexType.Z)))
            ):
            m.append((w,v))
            candidates.difference_update(g.neighbors(v))
            candidates.difference_update(g.neighbors(w))
    return m


def pauli_push(g: BaseGraph[VT,ET],
               matches: List[Tuple[VT,VT]]
               ) -> rules.RewriteOutputType[VT,ET]:
    """Pushes a Pauli (i.e. a pi phase) through another spider."""
    rem_verts: List[VT] = []
    rem_edges: List[ET] = []
    etab: Dict[Tuple[VT,VT], List[int]] = dict()
    for w,v in matches:  # w is a Pauli and v is the spider we are gonna push it through
        if g.vertex_degree(w) == 2:
            rem_verts.append(w)
            l = list(g.neighbors(w))
            l.remove(v)
            v2 = l[0]
            et1 = g.edge_type(g.edge(v,w))
            et2 = g.edge_type(g.edge(v2,w))
            etab[upair(v,v2)] = [1,0] if et1 == et2 else [0,1]
        else:
            g.set_phase(w,0)

        new_verts = []
        if vertex_is_zx(g.type(v)):
            g.scalar.add_phase(g.phase(v))
            g.set_phase(v,(-g.phase(v)) % 2)
            t = toggle_vertex(g.type(v))
            p: FractionLike = Fraction(1)
        else:
            t = VertexType.Z
            p = 0
        for edge in g.incident_edges(v):
            st = g.edge_st(edge)
            n = st[0] if st[1] == v else st[1]
            if n == w: continue
            r = 0.5*(g.row(n) + g.row(v))
            q = 0.5*(g.qubit(n) + g.qubit(v))
            et = g.edge_type(edge)
            rem_edges.append(edge)
            w2 = g.add_vertex(t,q,r,p)
            etab[upair(v,w2)] = [1,0]
            etab[upair(n,w2)] = [1,0] if et == EdgeType.SIMPLE else [0,1]
            new_verts.append(w2)
        if not vertex_is_zx(g.type(v)): # v is H_BOX
            if len(new_verts) == 2:
                etab[upair(new_verts[0],new_verts[1])] = [0,1]
            else:
                r = (g.row(v) + sum(g.row(n) for n in new_verts)) / (len(new_verts) + 1)
                q = (g.qubit(v) + sum(g.qubit(n) for n in new_verts))/(len(new_verts)+1)
                h = g.add_vertex(VertexType.H_BOX,q,r,Fraction(1))
                for n in new_verts: etab[upair(h,n)] = [1,0]
    return (etab, rem_verts, rem_edges, False)


def match_hadamard_edge(
        g: BaseGraph[VT,ET], 
        edgef: Optional[Callable[[ET],bool]] = None
        ) -> List[ET]:
    if edgef is not None: candidates = set([e for e in g.edges() if edgef(e)])
    else: candidates = g.edge_set()
    return [e for e in candidates if g.edge_type(e)==EdgeType.HADAMARD]


def match_edge(
        g: BaseGraph[VT,ET], 
        edgef: Optional[Callable[[ET],bool]] = None
        ) -> List[ET]:
    if edgef is not None: candidates = set([e for e in g.edges() if edgef(e)])
    else: candidates = g.edge_set()
    return list(candidates)


def euler_expansion(g: BaseGraph[VT,ET], 
                    matches: List[ET]
                    ) -> rules.RewriteOutputType[VT,ET]:
    """Expands the given Hadamard-edges into pi/2 phases using its Euler decomposition."""
    types = g.types()
    phases = g.phases()
    rem_edges = []
    etab = {}
    for e in matches:
        rem_edges.append(e)
        v1,v2 = g.edge_st(e)
        if vertex_is_zx(types[v1]) and types[v1] == types[v2]:
            r = 0.5*(g.row(v1) + g.row(v2))
            q = 0.5*(g.qubit(v1) + g.qubit(v2))
            t = toggle_vertex(types[v1])
            v = g.add_vertex(t,q,r)
            etab[upair(v,v1)] = [1,0]
            etab[upair(v,v2)] = [1,0]
            if phases[v1] == Fraction(1,2) or phases[v2] == Fraction(1,2):
                g.add_to_phase(v1,Fraction(3,2))
                g.add_to_phase(v2,Fraction(3,2))
                g.set_phase(v, Fraction(3,2))
                g.scalar.add_phase(Fraction(1,4))
            else:
                g.add_to_phase(v1,Fraction(1,2))
                g.add_to_phase(v2,Fraction(1,2))
                g.set_phase(v, Fraction(1,2))
                g.scalar.add_phase(Fraction(7,4))
        else:
            r = 0.25*g.row(v1) + 0.75*g.row(v2)
            q = 0.25*g.qubit(v1) + 0.75*g.qubit(v2)
            w1 = g.add_vertex(VertexType.Z,q,r,Fraction(1,2))
            etab[upair(v2,w1)] = [1,0]
            r = 0.5*g.row(v1) + 0.5*g.row(v2)
            q = 0.5*g.qubit(v1) + 0.5*g.qubit(v2)
            w2 = g.add_vertex(VertexType.X,q,r,Fraction(1,2))
            etab[upair(w1,w2)] = [1,0]
            r = 0.75*g.row(v1) + 0.25*g.row(v2)
            q = 0.75*g.qubit(v1) + 0.25*g.qubit(v2)
            w3 = g.add_vertex(VertexType.Z,q,r,Fraction(1,2))
            etab[upair(w2,w3)] = [1,0]
            etab[upair(w3,v1)] = [1,0]
            g.scalar.add_phase(Fraction(7,4))
            
    return (etab, [], rem_edges, False)

def add_Z_identity(g: BaseGraph[VT,ET], 
        matches: List[ET]
        ) -> rules.RewriteOutputType[VT,ET]:
    rem_edges = []
    etab = {}
    for e in matches:
        rem_edges.append(e)
        et = g.edge_type(e)
        v1,v2 = g.edge_st(e)
        r = 0.5*(g.row(v1) + g.row(v2))
        q = 0.5*(g.qubit(v1) + g.qubit(v2))
        w = g.add_vertex(VertexType.Z, q,r, 0)
        etab[upair(v1,w)] = [1,0] if et == EdgeType.SIMPLE else [0,1]
        etab[upair(v2,w)] = [1,0]
    return (etab, [], rem_edges, False)

def match_bialgebra(g: BaseGraph[VT,ET], 
        edgef: Optional[Callable[[ET],bool]] = None
        ) -> List[Tuple[VT,VT]]:
    if edgef is not None: candidates = set([e for e in g.edges() if edgef(e)])
    else: candidates = g.edge_set()
    m = []
    types = g.types()
    phases = g.phases()
    while len(candidates) > 0:
        e = candidates.pop()
        if g.edge_type(e) != EdgeType.SIMPLE: continue
        v,w = g.edge_st(e)
        if types[v] != VertexType.X:
            v,w = w,v
        if types[v] != VertexType.X: continue
        if types[w] == VertexType.Z:
            if phases[v] != 0 or phases[w] != 0: continue
            m.append((v,w))
            for n in g.neighbors(v):
                candidates.difference_update(g.incident_edges(n))
            for n in g.neighbors(w):
                candidates.difference_update(g.incident_edges(n))
        if types[w] == VertexType.H_BOX:
            if phases[v] != 0 or phases[w] != 1: continue
            m.append((v,w))
            for n in g.neighbors(v):
                candidates.difference_update(g.incident_edges(n))
            for n in g.neighbors(w):
                candidates.difference_update(g.incident_edges(n))
    return m

def bialgebra(g: BaseGraph[VT,ET],
        matches: List[Tuple[VT,VT]]
        ) -> rules.RewriteOutputType[VT,ET]:
    rem_verts = []
    etab = {}
    for v1, v2 in matches:
        rem_verts.append(v1)
        rem_verts.append(v2)
        v = (v1,v2)
        new_verts: Tuple[List[VT],List[VT]] = ([],[]) # new vertices for v1 and v2

        for i, j in [(0, 1), (1, 0)]:
            multi_edge_found = False
            for e in g.incident_edges(v[i]):
                source, target = g.edge_st(e)
                other_vertex = source if source != v[i] else target
                if other_vertex != v[j] or multi_edge_found:
                    q = 0.4*g.qubit(other_vertex) + 0.6*g.qubit(v[i])
                    r = 0.4*g.row(other_vertex) + 0.6*g.row(v[i])
                    newv = g.add_vertex(g.type(v[j]), qubit=q, row=r)
                    g.set_phase(newv, g.phase(v[j]))
                    new_verts[i].append(newv)
                    if other_vertex == v[j]:
                        q = 0.4*g.qubit(v[i]) + 0.6*g.qubit(other_vertex)
                        r = 0.4*g.row(v[i]) + 0.6*g.row(other_vertex)
                        newv2 = g.add_vertex(g.type(v[i]), qubit=q, row=r)
                        new_verts[j].append(newv2)
                        other_vertex = newv2
                    if upair(newv, other_vertex) not in etab:
                        etab[upair(newv, other_vertex)] = [0, 0]
                    type_index = 0 if g.edge_type(e) == EdgeType.SIMPLE else 1
                    etab[upair(newv, other_vertex)][type_index] += 1
                elif i == 0: # only add new vertex once
                    multi_edge_found = True

        for n1 in new_verts[0]:
            for n2 in new_verts[1]:
                if upair(n1,n2) not in etab:
                    etab[upair(n1,n2)] = [0, 0]
                etab[upair(n1,n2)][0] += 1

        if g.type(v1) == VertexType.H_BOX or g.type(v2) == VertexType.H_BOX: # x-h bialgebra
            x_vertex = v1 if g.type(v2) == VertexType.H_BOX else v2
            g.scalar.add_power(g.vertex_degree(x_vertex)-2)
        else: # z-x bialgebra
            g.scalar.add_power((g.vertex_degree(v1)-2)*(g.vertex_degree(v2)-2))
    return (etab, rem_verts, [], False)

def match_bialgebra_op(g: BaseGraph[VT,ET],
        vertexf: Optional[Callable[[VT], bool]] = None,
        vertex_type: Optional[Tuple[VertexType, VertexType]] = None,
        edge_type: Optional[EdgeType] = None
        ) -> Optional[Tuple[List[VT], List[VT]]]:
    if vertexf is not None: candidates = set([v for v in g.vertices() if vertexf(v)])
    else: candidates = g.vertex_set()
    if vertex_type is not None:
        vtype1, vtype2 = vertex_type
    else:
        vtype1, vtype2 = VertexType.Z, VertexType.X
    if edge_type is None:
        edge_type = EdgeType.SIMPLE
    type1_vertices = [v for v in candidates if g.type(v) == vtype1]
    type2_vertices = [v for v in candidates if g.type(v) == vtype2]
    if len(type1_vertices) <= 1 or len(type2_vertices) <= 1:
        return None
    #the vertices must be phase-free
    for v in type1_vertices + type2_vertices:
        if g.phase(v) != 0:
            return None
    # the vertices must have one external edge
    for v1 in type1_vertices:
        if g.vertex_degree(v1) != len(type2_vertices) + 1:
            return None
    for v2 in type2_vertices:
        if g.vertex_degree(v2) != len(type1_vertices) + 1:
            return None
    # if all type1 vertices are connected to all type2 vertices with a simple edge, then they are a match
    for v1 in type1_vertices:
        for v2 in type2_vertices:
            edges = list(g.edges(v1, v2))
            if not (len(edges) == 1 and g.edge_type(edges[0]) == edge_type):
                return None
    return type1_vertices, type2_vertices

def bialgebra_op(g: BaseGraph[VT,ET],
        matches: Tuple[List[VT], List[VT]],
        edge_type: Optional[EdgeType] = EdgeType.SIMPLE
        ) -> rules.RewriteOutputType[VT,ET]:
    def get_neighbors_and_loops(type1_vertices: List[VT], type2_vertices: List[VT]) -> Tuple[List[Tuple[VT, EdgeType]], List[EdgeType]]:
        neighbors: List[Tuple[VT, EdgeType]] = []
        loops: List[EdgeType] = []
        for v1 in type1_vertices:
            for edge in g.incident_edges(v1):
                edge_st = g.edge_st(edge)
                neighbor = edge_st[0] if edge_st[0] != v1 else edge_st[1]
                if neighbor in type2_vertices:
                    continue
                elif neighbor in type1_vertices:
                    if v1 > neighbor:
                        loops.append(g.edge_type(edge))
                else:
                    neighbors.append((neighbor, g.edge_type(edge)))
        return neighbors, loops

    def add_vertex_with_averages(vertices, g, vtype):
        average_row = sum(g.row(v) for v in vertices) / len(vertices)
        average_qubit = sum(g.qubit(v) for v in vertices) / len(vertices)
        return g.add_vertex(vtype, average_qubit, average_row)

    def update_etab(etab, new_vertex, neighbors, loops):
        for n, et in neighbors + [(new_vertex, et) for et in loops]:
            etab[upair(new_vertex, n)][0 if et == EdgeType.SIMPLE else 1] += 1

    type1_vertices, type2_vertices = matches
    neighbors1, loops1 = get_neighbors_and_loops(type1_vertices, type2_vertices)
    neighbors2, loops2 = get_neighbors_and_loops(type2_vertices, type1_vertices)

    new_vertex1 = add_vertex_with_averages(type1_vertices, g, g.type(type2_vertices[0]))
    new_vertex2 = add_vertex_with_averages(type2_vertices, g, g.type(type1_vertices[0]))

    etab: dict = defaultdict(lambda: [0, 0])
    if edge_type == EdgeType.SIMPLE:
        etab[upair(new_vertex1, new_vertex2)] = [1, 0]
    else:
        etab[upair(new_vertex1, new_vertex2)] = [0, 1]
    update_etab(etab, new_vertex1, neighbors1, loops1)
    update_etab(etab, new_vertex2, neighbors2, loops2)

    g.scalar.add_power(-(len(neighbors1)-1)*(len(neighbors2)-1))

    return (etab, type1_vertices + type2_vertices, [], False)


MATCHES_VERTICES = 1
MATCHES_EDGES = 2

operations = {
    "spider": {"text": "fuse spiders", 
               "tooltip": "Fuses connected spiders of the same color",
               "matcher": rules.match_spider_parallel, 
               "rule": rules.spider, 
               "type": MATCHES_VERTICES},
    "to_z": {"text": "change color to Z", 
               "tooltip": "Changes X spiders into Z spiders by pushing out Hadamards",
               "matcher": match_X_spiders, 
               "rule": color_change, 
               "type": MATCHES_VERTICES},
    "to_x": {"text": "change color to X", 
               "tooltip": "Changes Z spiders into X spiders by pushing out Hadamards",
               "matcher": match_Z_spiders, 
               "rule": color_change, 
               "type": MATCHES_VERTICES},
    "rem_id": {"text": "remove identity", 
               "tooltip": "Removes a 2-ary phaseless spider",
               "matcher": rules.match_ids_parallel, 
               "rule": rules.remove_ids, 
               "type": MATCHES_VERTICES},
    "id_z": {"text": "Add Z identity", 
               "tooltip": "Adds a phaseless arity 2 Z spider on the selected edges",
               "matcher": match_edge, 
               "rule": add_Z_identity, 
               "type": MATCHES_EDGES},
    "z_to_z_box": {"text": "Convert Z-spider to Z-box",
                "tooltip": "Converts a Z-spider into a Z-box",
                "matcher": rules.match_z_to_z_box_parallel,
                "rule": rules.z_to_z_box,
                "type": MATCHES_VERTICES},
    "hopf": {"text": "Hopf rule",
             "tooltip": "Applies the Hopf rule between pairs of spiders that share parallel edges",
             "matcher": rules.match_hopf,
             "rule": rules.hopf,
             "type": MATCHES_VERTICES},
    "remove_self_loops": {"text": "remove self-loops",
                        "tooltip": "Removes all self-loops on a spider",
                        "matcher": rules.match_self_loop,
                        "rule": rules.remove_self_loops,
                        "type": MATCHES_VERTICES},
    "had2edge": {"text": "Convert H-box", 
               "tooltip": "Converts an arity 2 H-box into an H-edge.",
               "matcher": hrules.match_hadamards,
               "rule": hrules.hadamard_to_h_edge,
               "type": MATCHES_VERTICES},
    "fuse_hbox": {"text": "Fuse H-boxes", 
               "tooltip": "Merges two adjacent H-boxes together",
               "matcher": hrules.match_connected_hboxes,
               "rule": hrules.fuse_hboxes,
               "type": MATCHES_EDGES},
    "mult_hbox": {"text": "Multiply H-boxes", 
               "tooltip": "Merges groups of H-boxes that have the same connectivity",
               "matcher": hrules.match_par_hbox,
               "rule": hrules.par_hbox,
               "type": MATCHES_VERTICES},
    "fuse_w": {"text": "fuse W nodes",
               "tooltip": "Merges two connected W nodes together",
               "matcher": rules.match_w_fusion_parallel,
                "rule": rules.w_fusion,
                "type": MATCHES_EDGES},
    "copy": {"text": "copy 0/pi spider", 
               "tooltip": "Copies a single-legged spider with a 0/pi phase through its neighbor",
               "matcher": hrules.match_copy,
               "rule": hrules.apply_copy,
               "type": MATCHES_VERTICES},
    "pauli": {"text": "push Pauli", 
               "tooltip": "Pushes an arity 2 pi-phase through a selected neighbor",
               "matcher": pauli_matcher, 
               "rule": pauli_push, 
               "type": MATCHES_VERTICES},
    "bialgebra": {"text": "bialgebra",
               "tooltip": "Applies the bialgebra rule to a connected pair of Z and X spiders",
               "matcher": match_bialgebra,
               "rule": bialgebra,
               "type": MATCHES_EDGES},
    "bialgebra_op": {"text": "bialgebra (inverse)",
               "tooltip": "Applies the bialgebra rule to a connected pair of Z and X spiders in the opposite direction",
               "matcher": match_bialgebra_op,
               "rule": bialgebra_op,
               "type": MATCHES_VERTICES},
    "euler": {"text": "decompose Hadamard", 
               "tooltip": "Expands a Hadamard-edge into its component spiders using its Euler decomposition",
               "matcher": match_hadamard_edge, 
               "rule": euler_expansion, 
               "type": MATCHES_EDGES},
    "lcomp": {"text": "local complementation", 
               "tooltip": "Deletes a spider with a pi/2 phase by performing a local complementation on its neighbors",
               "matcher": rules.match_lcomp_parallel, 
               "rule": rules.lcomp, 
               "type": MATCHES_VERTICES},
    "pivot": {"text": "pivot", 
               "tooltip": "Deletes a pair of spiders with 0/pi phases by performing a pivot",
               "matcher": lambda g, matchf: rules.match_pivot_parallel(g, matchf, check_edge_types=True), 
               "rule": rules.pivot, 
               "type": MATCHES_EDGES}
}


def operations_to_js() -> str:
    global operations
    return json.dumps({k:
            {"active":False, 
             "text":v["text"], 
             "tooltip":v["tooltip"]
            } for k,v in operations.items()})



# PyZX - Python library for quantum circuit rewriting
#        and optimisation using the ZX-calculus
# Copyright (C) 2019 - Aleks Kissinger, John van de Wetering,
#                      and Arianne Meijer-van de Griend

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from enum import Enum
import sys, os
import time
import argparse
from typing import Any, Dict, Iterable, List, Union

from pyzx_param.routing.cnot_mapper import ElimMode, gauss, genetic_elim_modes, basic_elim_modes, pso_elim_modes, sequential_gauss, elim_modes
from pyzx_param.routing.parity_maps import CNOT_tracker

try:
    from pandas import DataFrame
except:
    DataFrame = None
    if __name__ == "__main__":
        print("Warning: could not import pandas. No performance data will be exported.")
sys.path.append("..")
import numpy as np


from ..linalg import CNOTMaker, Mat2
from ..graph import Graph
from ..simplify import teleport_reduce, interior_clifford_simp
from ..circuit import Circuit
from ..optimize import basic_optimization
from ..routing.architecture import (
    Architecture,
    architectures,
    SQUARE,
    dynamic_size_architectures,
    create_architecture,
)
from ..routing.machine_learning import ParticleSwarmOptimization
from ..drawing import draw
from ..utils import make_into_list, restricted_float

description = (
    "Compiles given qasm files or those in the given folder to a given architecture."
)

debug = False

from ..extract import bi_adj, connectivity_from_biadj, permutation_as_swaps

class CompileMode(Enum):
    """
    Compilation modes for the cnot mapper procedures
    """

    QUIL_COMPILER = "quilc"
    NO_COMPILER = "not_compiled"
    TKET_COMPILER = "tket"

    def __str__(self):
        return f"{self.value}"

def create_dest_filename(
    original_file,
    outformat,
    population=None,
    iteration=None,
    crossover_prob=None,
    mutation_prob=None,
    index=None,
    pso_size=None,
    pso_step=None,
    pso_sco=None,
    pso_pco=None,
    pso_m=None,
):
    pop_ext = "" if population is None else "pop" + str(population)
    iter_ext = "" if iteration is None else "iter" + str(iteration)
    crossover_ext = "" if crossover_prob is None else "crossover" + str(crossover_prob)
    mutation_ext = "" if mutation_prob is None else "mutate" + str(mutation_prob)
    pso_ext = (
        "pso"
        + "_".join(
            [
                s + str(v)
                for s, v in zip(
                    ["swarm", "steps", "sco", "pco", "mut"],
                    [pso_size, pso_step, pso_sco, pso_pco, pso_m],
                )
            ]
        )
        if pso_size != None
        else ""
    )
    index_ext = "" if index is None else "(" + str(index) + ")"
    filename = os.path.basename(original_file)
    base_file, extension = os.path.splitext(filename)
    if outformat not in ("qasm", "qc", "quipper", "match"):
        print(
            "Unsupported circuit type {}. Please use qasm, qc or quipper".format(
                outformat
            )
        )
        return
    elif outformat != "match":
        extension = outformat

    new_filename = (
        "_".join(
            [
                part
                for part in [
                    base_file,
                    pop_ext,
                    iter_ext,
                    crossover_ext,
                    mutation_ext,
                    pso_ext,
                    index_ext,
                ]
                if part != ""
            ]
        )
        + extension
    )
    return new_filename


def read_circuit(source):
    if not os.path.exists(source):
        print("File {} does not exist".format(source))
        return
    return Circuit.load(source)


def simple_extract_no_gadgets(
    g,
    extract_cnots=None,
    quiet=True,
    initial_qubit_placement=None,
    allow_output_perm=False,
):
    if initial_qubit_placement is None:
        initial_qubit_placement = [i for i in range(g.qubit_count())]
    g.normalise()
    g = preprocess_graph(g)
    qs = g.qubits()  # We are assuming that these are objects that update...
    rs = g.rows()  # ...to reflect changes to the graph, so that when...
    ty = (
        g.types()
    )  # ... g.set_row/g.set_qubit is called, these things update directly to reflect that
    phases = g.phases()
    if extract_cnots is None:
        extract_cnots = lambda mat2, p, col=False: Mat2(
            [[mat2.data[i][j] for j in p] if col else mat2.data[i] for i in p]
        ).to_cnots()

    h = Graph()

    qindex = {}
    depth = 0
    for i in range(len(g.inputs)):
        v = h.add_vertex(0, i, depth)
        h.inputs.append(v)
        qindex[i] = v
    depth = 1

    def add_phase_gate(q, phase):
        nonlocal depth
        v = h.add_vertex(1, q, depth, phase)
        h.add_edge((qindex[q], v), 1)
        qindex[q] = v
        depth += 1
        return v

    def add_hadamard(q):
        nonlocal depth
        v = h.add_vertex(1, q, depth)
        h.add_edge((qindex[q], v), 2)
        qindex[q] = v
        depth += 1
        return v

    def add_cnot(ctrl, tgt):
        nonlocal depth
        v1 = h.add_vertex(1, ctrl, depth)
        v2 = h.add_vertex(2, tgt, depth)
        h.add_edges([(qindex[ctrl], v1), (qindex[tgt], v2), (v1, v2)], 1)
        qindex[ctrl] = v1
        qindex[tgt] = v2
        depth += 1
        return v1, v2

    leftrow = 1

    nodes = []  # Non phase-gadgets
    for v in g.vertices():  # Find which vertices are gadgets
        if rs[v] > 1:
            g.set_row(v, rs[v] + 20)
        if v in g.inputs or v in g.outputs:
            continue
        elif all(
            w in g.inputs or w in g.outputs or len(list(g.neighbours(w))) != 1
            for w in g.neighbours(v)
        ):  # regular vertex
            nodes.append(v)
    nodestotal = 19

    processed_targets = {}
    while True:
        left = [v for v in g.vertices() if rs[v] == leftrow]
        # fig = draw(g)
        # fig.savefig("test.png", dpi=720)
        # import matplotlib.pyplot as plt
        # plt.close(fig)
        for v in left:
            # First we add the gates to the circuit that can be processed now,
            # and we simplify the graph to represent this.
            q = qs[v]
            phase = phases[v]
            t = ty[v]
            if t != 1:
                raise TypeError("Only supports zx-diagrams in graph-like state")
            neigh = [w for w in g.neighbours(v) if rs[w] < leftrow]
            if len(neigh) != 1:
                raise TypeError("Graph doesn't seem circuit like: multiple parents")
            n = neigh[0]
            if qs[n] != q:
                raise TypeError(
                    "Graph doesn't seem circuit like: cross qubit connections"
                )
            if g.edge_type(g.edge(n, v)) == 2:
                add_hadamard(q)
                g.set_edge_type(g.edge(n, v), 1)
            # if t == 0: continue # it is an output
            if phase != 0:
                add_phase_gate(q, phase)
                g.set_phase(v, 0)

        neighbours = set()
        finished = set()
        for v in left.copy():  # Deal with frontier connected to outputs
            d = [w for w in g.neighbours(v) if rs[w] > leftrow]
            if any(w in g.outputs for w in d):
                if len(d) == 1:
                    finished.add(v)
                b = [w for w in d if w in g.outputs][0]
                q = qs[b]
                r = rs[b]
                w = g.add_vertex(1, q, r - 1)
                nodes.append(w)
                e = g.edge(v, b)
                et = g.edge_type(e)
                g.remove_edge(e)
                g.add_edge((v, w), 2)
                g.add_edge((w, b), 3 - et)
                d.remove(b)
                d.append(w)
            neighbours.update(d)

        # if not left: break  # We are done
        if len(finished) == g.qubit_count() or not left:
            break
        left.sort(key=lambda v: g.qubit(v))
        right = [
            w for w in neighbours if w in nodes
        ]  # Only get non-phase-gadget neighbours
        m = bi_adj(g, right, left)
        cnots = extract_cnots(m, initial_qubit_placement)
        for cnot in cnots:
            m.row_add(cnot.target, cnot.control)
            add_cnot(qs[left[cnot.control]], qs[left[cnot.target]])
        connectivity_from_biadj(g, m, right, left)
        good_verts = {}
        for i, row in enumerate(m.data):
            if sum(row) == 1:
                v = left[i]
                w = right[[j for j in range(len(m.data[i])) if m.data[i][j]][0]]
                is_connected = lambda v, l: any([g.connected(v, w) for w in l])
                if not is_connected(w, good_verts.values()):
                    good_verts[v] = w
        if not good_verts:
            print(m)
            print(left)
            print(right)
            print(nodes)
            raise Exception("No good match found")
        for v in left:
            if v not in good_verts:
                g.set_row(v, leftrow + 1)
            else:
                g.set_row(good_verts[v], leftrow + 1)
                g.set_qubit(good_verts[v], qs[v])
                if len(list(g.neighbours(v))) > 2:  # Gadgets are still connected to it
                    w = add_phase_gate(qs[v], 0)
                    processed_targets[v] = w
        leftrow += 1
        if leftrow >= nodestotal:
            nodestotal += 20
            for v in g.vertices():
                if rs[v] > leftrow:
                    g.set_row(v, rs[v] + 20)

    # We are done processing now. Time to deal with swaps.
    left.sort(key=lambda v: g.qubit(v))
    right = [w for v in left for w in g.neighbours(v) if rs[w] > leftrow]
    if not allow_output_perm:
        right.sort(key=lambda v: g.qubit(v))
        m = bi_adj(g, right, left)
        cnots = extract_cnots(m, initial_qubit_placement, col=True)  # Does Gauss
        for cnot in cnots:
            m.row_add(cnot.target, cnot.control)
            add_cnot(cnot.control, cnot.target)
        connectivity_from_biadj(g, m, right, left)  # Removes connections from g

        for i, w in enumerate(g.outputs):
            n = list(g.neighbours(w))[0]
            et = g.edge_type(g.edge(n, w))
            v = h.add_vertex(0, i, depth)
            h.outputs.append(v)
            h.add_edge((qindex[i], v), 3 - et)
            qindex[i] = v
        final_placement = [i for i in range(g.qubit_count())]
    else:
        final_placement = [g.qubit(r) for r in right]
        for q, l in enumerate(left):
            n = right[q]
            r = [v for v in g.outputs if n == list(g.neighbours(v))[0]][0]
            et = g.edge_type(g.edge(n, r))
            v = h.add_vertex(0, q, depth)
            h.outputs.append(v)
            h.add_edge((qindex[q], v), 3 - et)
            qindex[q] = v
    return h, final_placement


def preprocess_graph(g):
    g.normalise()
    for v in g.vertices():  # increase the row
        if v not in g.inputs:
            r = g.row(v)
            g.set_row(v, r + 2)
    for input in g.inputs:
        n = list(g.neighbours(input))[0]
        q = g.qubit(input)
        edge = g.edge(input, n)
        etype = g.edge_type(edge)
        v1 = g.add_vertex(1, q, 1)
        g.add_edge((input, v1), edgetype=1)
        if etype == 1:  # Normal edge -> add -o-H-o-H-
            v2 = g.add_vertex(1, q, 2)
            g.add_edge((v1, v2), edgetype=2)
            v1 = v2
            # else Hadamard edge -> add -o-
        g.add_edge((v1, n), edgetype=2)
        g.remove_edge(edge)
    for output in g.outputs:
        n = list(g.neighbours(output))[0]
        edge = g.edge(output, n)
        etype = g.edge_type(edge)
        r = g.row(output)
        g.set_row(output, r + 1)
        if etype == 2:  # Normal edge -> append -o-
            q = g.qubit(output)
            v1 = g.add_vertex(1, q, r)
            g.add_edge((n, v1), edgetype=2)
            g.add_edge((v1, output), edgetype=1)
            g.remove_edge(edge)
    g.normalise()
    return g


def route_circuit(
    c,
    architecture,
    mode=ElimMode.STEINER_MODE,
    dest_file=None,
    population=30,
    iterations=15,
    crossover_prob=0.8,
    mutation_prob=0.2,
    allow_output_perm=True,
    swarm_size=15,
    n_steps=10,
    s_crossover=0.4,
    p_crossover=0.3,
    pso_mutation=0.2,
):
    metric = lambda c: len(
        [
            gate
            for gate in c.gates
            if hasattr(gate, "name") and gate.name in ["CNOT", "CZ"]
        ]
    )
    if mode == CompileMode.QUIL_COMPILER:
        from pyzx_param.pyquil_circuit import PyQuilCircuit

        compiled_circuit = PyQuilCircuit.from_circuit(
            c, architecture
        )  # TODO fix this to include other gates aswell
        compiled_circuit.compile()
    else:
        g = c.to_graph()
        g = teleport_reduce(g)
        interior_clifford_simp(g)
        g = g.copy()  # reduces the number of gates when extracting.
        if type(architecture) == type(""):
            architecture = create_architecture(architecture)

        def gauss_func(m, permutation=None, col=False):
            cn = CNOTMaker()
            if False:  # mode == ElimMode.GAUSS_MODE:
                return m.to_cnots()
            if col:
                if permutation is None:
                    permutation = list(range(m.cols()))
                m = Mat2([[row[col] for col in permutation] for row in m.data])
                rank = gauss(
                    ElimMode.STEINER_MODE,
                    m,
                    architecture,
                    full_reduce=True,
                    x=cn,
                    permutation=permutation,
                )
            else:
                m = m.copy()
                if mode in genetic_elim_modes:
                    rank = gauss(
                        mode,
                        m,
                        architecture,
                        full_reduce=True,
                        x=cn,
                        permutation=permutation,
                        row=False,
                        col=True,
                        population_size=population,
                        crossover_prob=crossover_prob,
                        mutate_prob=mutation_prob,
                        n_iterations=iterations,
                    )
                else:
                    rank = gauss(
                        mode,
                        m,
                        architecture,
                        full_reduce=True,
                        x=cn,
                        permutation=permutation,
                    )
            # print(permutation)
            # print(len(cn.cnots), cn.cnots)
            return cn.cnots

        best_permutation = [i for i in range(g.qubit_count())]
        best_solution = None
        if allow_output_perm not in [True, False]:  # mode in genetic_elim_modes:

            def fitness(permutation):
                new_g = g.copy()
                compiled_g, _ = simple_extract_no_gadgets(
                    new_g,
                    gauss_func,
                    initial_qubit_placement=permutation.tolist(),
                    allow_output_perm=allow_output_perm,
                )
                compiled_circuit = Circuit.from_graph(compiled_g)
                compiled_circuit = basic_optimization(compiled_circuit, do_swaps=False)
                return metric(compiled_circuit)

            # Build a reversed circuit to find a new initial qubit placement.
            rev_c = Circuit(g.qubit_count())
            for gate in reversed(c.gates):
                rev_c.add_gate(gate)
            rev_g = rev_c.to_graph()
            rev_g = teleport_reduce(rev_g)
            interior_clifford_simp(rev_g)
            rev_g = rev_g.copy()  # reduces the number of gates when extracting.

            def step(permutation):
                new_g = g.copy()
                compiled_g, output_perm = simple_extract_no_gadgets(
                    new_g,
                    gauss_func,
                    initial_qubit_placement=permutation.tolist(),
                    allow_output_perm=allow_output_perm,
                )
                compiled_circuit = Circuit.from_graph(compiled_g)
                compiled_circuit = basic_optimization(compiled_circuit, do_swaps=False)
                solution = (permutation.tolist(), compiled_circuit, output_perm)
                new_g = rev_g.copy()
                compiled_g, output_perm = simple_extract_no_gadgets(
                    new_g,
                    gauss_func,
                    initial_qubit_placement=output_perm,
                    allow_output_perm=allow_output_perm,
                )
                # compiled_circuit = Circuit.from_graph(compiled_g)
                # compiled_circuit = basic_optimization(compiled_circuit, do_swaps=False)
                return np.asarray(output_perm), solution, metric(compiled_circuit)

            # optimizer = GeneticAlgorithm(population, crossover_prob, mutation_prob, fitness, quiet=False)
            optimizer = ParticleSwarmOptimization(
                swarm_size=swarm_size,
                step_func=step,
                s_best_crossover=s_crossover,
                p_best_crossover=p_crossover,
                mutation=pso_mutation,
            )
            best_solution = optimizer.find_optimum(
                architecture.n_qubits, n_steps, False
            )

        if best_solution is None:
            new_g = g.copy()
            compiled_g, final_placement = simple_extract_no_gadgets(
                new_g,
                gauss_func,
                initial_qubit_placement=best_permutation,
                allow_output_perm=allow_output_perm,
            )
            compiled_circuit = Circuit.from_graph(compiled_g)
            compiled_circuit = basic_optimization(compiled_circuit, do_swaps=False)
        else:
            best_permutation, compiled_circuit, final_placement = best_solution
        print("Permutation found!", best_permutation)
        print("Done extracting!", final_placement)
        # sanity checks.
        from ..tensor import compare_tensors

        print("Initial CNOT/CZ count:", metric(c))
        if allow_output_perm:
            compiled_circuit2 = Circuit.from_graph(compiled_circuit.to_graph())
            swap_map = {i: final_placement[i] for i, q in enumerate(best_permutation)}
            for t1, t2 in permutation_as_swaps(swap_map):
                compiled_circuit2.add_gate("CNOT", control=t1, target=t2)
                compiled_circuit2.add_gate("CNOT", control=t2, target=t1)
                compiled_circuit2.add_gate("CNOT", control=t1, target=t2)
            print(
                "Extract circuit equals initial circuit? (with appended swaps)",
                compare_tensors(c, compiled_circuit2),
            )  # check extraction procedure
        else:
            print(
                "Extract circuit equals initial circuit? (without appended swaps)",
                compare_tensors(c, compiled_circuit),
            )  # check extraction procedure
        qubit_lookup = {
            i: best_permutation.index(i) for i in range(len(best_permutation))
        }
        illegal_gates = [
            gate
            for gate in compiled_circuit.gates
            if hasattr(gate, "name")
            and (gate.name == "CNOT" or gate.name == "CZ")
            and not (
                architecture.graph.connected(
                    qubit_lookup[gate.target], qubit_lookup[gate.control]
                )
                or architecture.graph.connected(
                    qubit_lookup[gate.control], qubit_lookup[gate.target]
                )
            )
        ]
        print("All CNOT/CZ gates allowed in the architecture?", len(illegal_gates) == 0)
        if illegal_gates:
            print("Which ones?", illegal_gates)

    print("Final CNOT/CZ count:", metric(compiled_circuit))

    if dest_file is not None:
        print("Saving the resulting circuit.")
        compiled_qasm = compiled_circuit.to_qasm()
        with open(dest_file, "w") as f:
            f.write(compiled_qasm)
    return compiled_circuit


def batch_route_circuits(
    source,
    modes,
    architectures,
    n_qubits=None,
    populations=30,
    iterations=15,
    crossover_probs=0.8,
    mutation_probs=0.5,
    dest_folder=None,
    metrics_file=None,
    n_compile=1,
    allow_output_perm=True,
    swarm_size=15,
    n_steps=10,
    s_crossover=0.4,
    p_crossover=0.3,
    pso_mutation=0.2,
):
    modes = make_into_list(modes)
    architectures = make_into_list(architectures)
    populations = make_into_list(populations)
    iterations = make_into_list(iterations)
    crossover_probs = make_into_list(crossover_probs)
    mutation_probs = make_into_list(mutation_probs)
    swarm_size = make_into_list(swarm_size)
    n_steps = make_into_list(n_steps)
    s_crossover = make_into_list(s_crossover)
    p_crossover = make_into_list(p_crossover)
    pso_mutation = make_into_list(pso_mutation)
    if allow_output_perm == "y":
        allow_output_perm = True
    elif allow_output_perm == "n":
        allow_output_perm = False

    if os.path.isfile(source):
        source, file = os.path.split(source)
        files = [file]
    else:
        files = [
            f for f in os.listdir(source) if os.path.isfile(os.path.join(source, f))
        ]

    if not os.path.exists(source):
        raise IOError("Folder does not exist: " + source)
    if dest_folder is None:
        dest_folder = source
    else:
        os.makedirs(dest_folder, exist_ok=True)

    arch_iter = []
    circuits = {}
    metrics = []
    for architecture in architectures:
        if architecture in dynamic_size_architectures:
            if n_qubits is None:
                raise KeyError(
                    "Number of qubits not specified for architecture" + architecture
                )
            else:
                n_qubits = make_into_list(n_qubits)
                arch_iter.extend(
                    [create_architecture(architecture, n_qubits=q) for q in n_qubits]
                )
        else:
            arch_iter.append(create_architecture(architecture))
    for architecture in arch_iter:
        circuits[architecture.name] = {}
        for mode in modes:
            if mode == CompileMode.QUIL_COMPILER:
                n_compile_list = range(n_compile)
            else:
                n_compile_list = [None]
            new_dest_folder = os.path.join(dest_folder, architecture.name, mode)
            os.makedirs(new_dest_folder, exist_ok=True)
            if mode in genetic_elim_modes:
                pop_iter = populations
                iter_iter = iterations
                crossover_iter = crossover_probs
                mutation_iter = mutation_probs
                circuits[architecture.name][mode] = {}
            else:
                if mode == CompileMode.QUIL_COMPILER:
                    circuits[architecture.name][mode] = []
                pop_iter = [None]
                iter_iter = [None]
                crossover_iter = [None]
                mutation_iter = [None]
            if allow_output_perm in [True, False]:
                swarm_size = [None]
                n_steps = [None]
                p_crossover = [None]
                s_crossover = [None]
                pso_mutation = [None]
            for population in pop_iter:
                for iteration in iter_iter:
                    for crossover_prob in crossover_iter:
                        for mutation_prob in mutation_iter:
                            for swarm in swarm_size:
                                for steps in n_steps:
                                    for pco in p_crossover:
                                        for sco in s_crossover:
                                            for pso_m in pso_mutation:
                                                for file in files:
                                                    origin_file = os.path.join(
                                                        source, file
                                                    )
                                                    original_circuit = read_circuit(
                                                        origin_file
                                                    )
                                                    if original_circuit:
                                                        for i in n_compile_list:
                                                            dest_filename = (
                                                                create_dest_filename(
                                                                    origin_file,
                                                                    "match",
                                                                    population,
                                                                    iteration,
                                                                    crossover_prob,
                                                                    mutation_prob,
                                                                    i,
                                                                    swarm,
                                                                    steps,
                                                                    sco,
                                                                    pco,
                                                                    pso_m,
                                                                )
                                                            )
                                                            dest_file = os.path.join(
                                                                dest_folder,
                                                                architecture.name,
                                                                mode,
                                                                dest_filename,
                                                            )
                                                            original_circuit = (
                                                                read_circuit(
                                                                    origin_file
                                                                )
                                                            )
                                                            try:
                                                                start_time = time.time()
                                                                circuit = route_circuit(
                                                                    original_circuit,
                                                                    architecture,
                                                                    mode=mode,
                                                                    dest_file=dest_file,
                                                                    population=population,
                                                                    iterations=iteration,
                                                                    crossover_prob=crossover_prob,
                                                                    mutation_prob=mutation_prob,
                                                                    allow_output_perm=allow_output_perm,
                                                                    swarm_size=swarm,
                                                                    n_steps=steps,
                                                                    s_crossover=sco,
                                                                    p_crossover=pco,
                                                                    pso_mutation=pso_m,
                                                                )
                                                                end_time = time.time()
                                                                if (
                                                                    metrics_file
                                                                    is not None
                                                                ):
                                                                    metrics.append(
                                                                        make_metrics(
                                                                            circuit,
                                                                            origin_file,
                                                                            architecture.name,
                                                                            mode,
                                                                            dest_file,
                                                                            population,
                                                                            iteration,
                                                                            crossover_prob,
                                                                            mutation_prob,
                                                                            end_time
                                                                            - start_time,
                                                                            i,
                                                                        )
                                                                    )
                                                                if (
                                                                    mode
                                                                    in genetic_elim_modes
                                                                ):
                                                                    circuits[
                                                                        architecture.name
                                                                    ][mode][
                                                                        (
                                                                            population,
                                                                            iteration,
                                                                            crossover_prob,
                                                                            mutation_prob,
                                                                        )
                                                                    ] = circuit
                                                                elif (
                                                                    mode
                                                                    == CompileMode.QUIL_COMPILER
                                                                ):
                                                                    circuits[
                                                                        architecture.name
                                                                    ][mode].append(
                                                                        circuit
                                                                    )
                                                                else:
                                                                    circuits[
                                                                        architecture.name
                                                                    ][mode] = circuit
                                                            except KeyError as e:  # Should only happen with quilc
                                                                if (
                                                                    mode
                                                                    == CompileMode.QUIL_COMPILER
                                                                ):
                                                                    print(
                                                                        "\033[31mCould not compile",
                                                                        origin_file,
                                                                        "into",
                                                                        dest_file,
                                                                        end="\033[0m\n",
                                                                    )
                                                                else:
                                                                    raise e

    if len(metrics) > 0 and DataFrame != None:
        df = DataFrame(metrics)
        if os.path.exists(metrics_file):  # append to the file - do not overwrite!
            df.to_csv(
                metrics_file,
                columns=get_metric_header(),
                header=False,
                index=False,
                mode="a",
            )
        else:
            df.to_csv(metrics_file, columns=get_metric_header(), index=False)
    return circuits

def count_cnots_circuit(mode, circuit, n_compile=1, store_circuit_as=None):
    count = -1
    if mode == CompileMode.QUIL_COMPILER:
        from pyzx_param.pyquil_circuit import PyQuilCircuit

        if isinstance(circuit, PyQuilCircuit):
            count = (
                sum([circuit.compiled_cnot_count() for i in range(n_compile)])
                / n_compile
            )
    elif mode == CompileMode.NO_COMPILER:
        count = circuit.count_cnots()
    if store_circuit_as is not None:
        with open(store_circuit_as, "w") as f:
            f.write(circuit.to_qasm())
    return count


def get_metric_header():
    metrics = CNOT_tracker.get_metric_names()
    return (
        [
            "id",
            "architecture",
            "mode",
            "index",
            "population",
            "n_iterations",
            "crossover",
            "mutation",
        ]
        + metrics
        + ["time", "destination_file"]
    )


def make_metrics(
    circuit,
    id,
    architecture_name,
    mode,
    dest_file=None,
    population=None,
    iteration=None,
    crossover_prob=None,
    mutation_prob=None,
    passed_time=None,
    index=None,
):
    result = circuit.gather_metrics()
    result["id"] = id
    result["mode"] = mode
    result["architecture"] = architecture_name
    result["population"] = population
    result["n_iterations"] = iteration
    result["crossover"] = crossover_prob
    result["mutation"] = mutation_prob
    result["time"] = passed_time
    result["index"] = index
    result["destination_file"] = dest_file
    return result


def batch_map_cnot_circuits(
    source,
    modes,
    architectures,
    n_qubits=None,
    populations: Union[int, List[int]] = 30,
    iterations=15,
    crossover_probs=0.8,
    mutation_probs=0.5,
    dest_folder=None,
    metrics_file=None,
    n_compile=1,
) -> Dict[Architecture, Dict[ElimMode, Any]]:
    modes = make_into_list(modes)
    architectures = make_into_list(architectures)
    populations = make_into_list(populations)
    iterations = make_into_list(iterations)
    crossover_probs = make_into_list(crossover_probs)
    mutation_probs = make_into_list(mutation_probs)

    if os.path.isfile(source):
        source, file = os.path.split(source)
        files = [file]
    else:
        files = [
            f for f in os.listdir(source) if os.path.isfile(os.path.join(source, f))
        ]

    if not os.path.exists(source):
        raise IOError("Folder does not exist: " + source)
    if dest_folder is None:
        dest_folder = source
    else:
        os.makedirs(dest_folder, exist_ok=True)

    arch_iter = []
    circuits: Dict[Architecture, Dict[ElimMode, Any]] = {}
    metrics = []
    for architecture in architectures:
        if architecture in dynamic_size_architectures:
            if n_qubits is None:
                raise KeyError(
                    "Number of qubits not specified for architecture" + architecture
                )
            else:
                n_qubits = make_into_list(n_qubits)
                arch_iter.extend(
                    [create_architecture(architecture, n_qubits=q) for q in n_qubits]
                )
        else:
            arch_iter.append(create_architecture(architecture))
    for architecture in arch_iter:
        circuits[architecture.name] = {}
        for mode in modes:
            n_compile_list: Iterable[int]
            if mode == CompileMode.QUIL_COMPILER:
                n_compile_list = range(n_compile)
            else:
                n_compile_list = [0]
            new_dest_folder = os.path.join(dest_folder, architecture.name, mode)
            os.makedirs(new_dest_folder, exist_ok=True)
            if mode in genetic_elim_modes:
                pop_iter = populations
                iter_iter = iterations
                crossover_iter = crossover_probs
                mutation_iter = mutation_probs
                circuits[architecture.name][mode] = {}
            else:
                if mode == CompileMode.QUIL_COMPILER:
                    circuits[architecture.name][mode] = []
                pop_iter = [0]
                iter_iter = [0]
                crossover_iter = [0]
                mutation_iter = [0]

            for population in pop_iter:
                for iteration in iter_iter:
                    for crossover_prob in crossover_iter:
                        for mutation_prob in mutation_iter:
                            for file in files:
                                if os.path.splitext(file)[1].lower() == ".qasm":
                                    origin_file = os.path.join(source, file)
                                    for i in n_compile_list:
                                        dest_filename = create_dest_filename(
                                            origin_file,
                                            population,
                                            iteration,
                                            crossover_prob,
                                            mutation_prob,
                                            i,
                                        )
                                        dest_file = os.path.join(
                                            dest_folder,
                                            architecture.name,
                                            mode,
                                            dest_filename,
                                        )
                                        try:
                                            start_time = time.time()
                                            circuit = map_cnot_circuit(
                                                origin_file,
                                                architecture,
                                                mode=mode,
                                                dest_file=dest_file,
                                                population=population,
                                                iterations=iteration,
                                                crossover_prob=crossover_prob,
                                                mutation_prob=mutation_prob,
                                            )
                                            end_time = time.time()
                                            if metrics_file is not None:
                                                metrics.append(
                                                    make_metrics(
                                                        circuit,
                                                        origin_file,
                                                        architecture.name,
                                                        mode,
                                                        dest_file,
                                                        population,
                                                        iteration,
                                                        crossover_prob,
                                                        mutation_prob,
                                                        end_time - start_time,
                                                        i,
                                                    )
                                                )
                                            if mode in genetic_elim_modes:
                                                circuits[architecture.name][mode][
                                                    (
                                                        population,
                                                        iteration,
                                                        crossover_prob,
                                                        mutation_prob,
                                                    )
                                                ] = circuit
                                            elif mode == CompileMode.QUIL_COMPILER:
                                                circuits[architecture.name][
                                                    mode
                                                ].append(circuit)
                                            else:
                                                circuits[architecture.name][
                                                    mode
                                                ] = circuit
                                        except KeyError as e:  # Should only happen with quilc
                                            if mode == CompileMode.QUIL_COMPILER:
                                                print(
                                                    "\033[31mCould not compile",
                                                    origin_file,
                                                    "into",
                                                    dest_file,
                                                    end="\033[0m\n",
                                                )
                                            else:
                                                raise e

    if metrics_file is not None and len(metrics) > 0 and DataFrame is not None:
        df = DataFrame(metrics)
        print("Average gate count:", df["n_cnots"].mean())
        print("Average gate depth:", df["depth"].mean())
        if os.path.exists(metrics_file):  # append to the file - do not overwrite!
            df.to_csv(
                metrics_file,
                columns=get_metric_header(),
                header=False,
                index=False,
                mode="a",
            )
        else:
            df.to_csv(metrics_file, columns=get_metric_header(), index=False)
    return circuits


def map_cnot_circuit(
    file,
    architecture,
    mode=ElimMode.GENETIC_STEINER_MODE,
    dest_file=None,
    population=30,
    iterations=15,
    crossover_prob=0.8,
    mutation_prob=0.2,
    **kwargs,
):
    if type(architecture) == type(""):
        architecture = create_architecture(architecture)
    circuit = CNOT_tracker.from_qasm_file(file)
    matrix = circuit.matrix
    compiled_circuit = CNOT_tracker(circuit.n_qubits)
    if mode in basic_elim_modes:
        rank = gauss(
            mode, matrix, architecture, full_reduce=True, y=compiled_circuit, **kwargs
        )
    elif mode in genetic_elim_modes:
        rank = gauss(
            mode,
            matrix,
            architecture,
            full_reduce=True,
            y=compiled_circuit,
            population_size=population,
            crossover_prob=crossover_prob,
            mutate_prob=mutation_prob,
            n_iterations=iterations,
            **kwargs,
        )
    elif mode == CompileMode.QUIL_COMPILER:
        from pyzx_param.pyquil_circuit import PyQuilCircuit

        compiled_circuit = PyQuilCircuit.from_CNOT_tracker(circuit, architecture)
        compiled_circuit.compile()

    if dest_file is not None:
        compiled_qasm = compiled_circuit.to_qasm()
        with open(dest_file, "w") as f:
            f.write(compiled_qasm)
    return compiled_circuit


def sequential_map_cnot_circuits(
    source: os.PathLike,
    modes: Union[ElimMode, List[ElimMode]],
    architectures: Union[Architecture, List[Architecture]],
    n_qubits: Union[int, List[int], None] = None,
    populations=30,
    iterations=15,
    crossover_probs=0.8,
    mutation_probs=0.5,
    dest_folder=None,
    metrics_file=None,
    n_compile=1,
) -> Dict[str, Dict[ElimMode, Any]]:
    modes = make_into_list(modes)
    architectures = make_into_list(architectures)
    populations = make_into_list(populations)
    iterations = make_into_list(iterations)
    crossover_probs = make_into_list(crossover_probs)
    mutation_probs = make_into_list(mutation_probs)

    if os.path.isfile(source):
        source, file = os.path.split(source)
        files = [file]
    else:
        files = [
            f for f in os.listdir(source) if os.path.isfile(os.path.join(source, f))
        ]

    if not os.path.exists(source):
        raise IOError(f"Folder does not exist: {source}")
    if dest_folder is None:
        dest_folder = source
    else:
        os.makedirs(dest_folder, exist_ok=True)

    arch_iter = []
    circuits: Dict[str, Dict[ElimMode, Any]] = {}
    for architecture in architectures:
        if architecture in dynamic_size_architectures:
            if n_qubits is None:
                raise KeyError(
                    f"Number of qubits not specified for architecture {architecture}"
                )
            else:
                n_qubits = make_into_list(n_qubits)
                arch_iter.extend(
                    [create_architecture(architecture, n_qubits=q) for q in n_qubits]
                )
        else:
            arch_iter.append(create_architecture(architecture))
    for architecture in arch_iter:
        circuits[architecture.name] = {}
        for mode in modes:
            n_compile_list: Iterable[int]
            if mode == CompileMode.QUIL_COMPILER:
                n_compile_list = range(n_compile)
            else:
                n_compile_list = [0]
            new_dest_folder = os.path.join(dest_folder, architecture.name, str(mode))
            os.makedirs(new_dest_folder, exist_ok=True)
            if mode in genetic_elim_modes or mode in pso_elim_modes:
                pop_iter = populations
                iter_iter = iterations
                crossover_iter = crossover_probs
                mutation_iter = mutation_probs
                circuits[architecture.name][mode] = {}
            else:
                if mode == CompileMode.QUIL_COMPILER:
                    circuits[architecture.name][mode] = []
                pop_iter = [None]
                iter_iter = [None]
                crossover_iter = [None]
                mutation_iter = [None]

            for population in pop_iter:
                for iteration in iter_iter:
                    for crossover_prob in crossover_iter:
                        for mutation_prob in mutation_iter:
                            matrices = []
                            origin_file = None
                            for file in files:
                                if os.path.splitext(file)[1].lower() == ".qasm":
                                    origin_file = os.path.join(source, file)
                                    matrices.append(
                                        CNOT_tracker.from_qasm_file(origin_file).matrix
                                    )
                            for i in n_compile_list:
                                # dest_filename = create_dest_filename(origin_file, population, iteration, crossover_prob, mutation_prob, i)
                                # dest_file = os.path.join(dest_folder, architecture.name, mode, dest_filename)
                                try:
                                    print("Sequence length", len(matrices))
                                    start_time = time.time()
                                    _circs, permutations, score = sequential_gauss(
                                        matrices,
                                        architecture=architecture,
                                        mode=mode,
                                        population_size=population,
                                        n_iterations=iteration,
                                        crossover_prob=crossover_prob,
                                        mutate_prob=mutation_prob,
                                    )
                                    end_time = time.time()
                                    print("score:", score)
                                    print("time (s):", end_time - start_time)
                                except KeyError as e:  # Should only happen with quilc
                                    if mode == CompileMode.QUIL_COMPILER:
                                        print(
                                            "\033[31mCould not compile",
                                            origin_file,
                                            end="\033[0m\n",
                                        )
                                    else:
                                        raise e

    return circuits

def main(args):
    parser = argparse.ArgumentParser(prog="pyzx router", description=description)
    parser.add_argument("QASM_source",nargs="+",help="The QASM file or folder with QASM files to be routed.")
    parser.add_argument("-m","--mode",nargs="+",dest="mode",default=ElimMode.STEINER_MODE,help="The mode specifying how to route. choose 'all' for using all modes.",choices=elim_modes + [CompileMode.QUIL_COMPILER, "all"])
    parser.add_argument("-a","--architecture",nargs="+",dest="architecture",default=SQUARE,choices=architectures,help="Which architecture it should run compile to.")
    parser.add_argument("-q","--qubits",nargs="+",default=None,type=int,help="The number of qubits for the architecture.")
    parser.add_argument("--output_perm",default="y",choices=["y", "n", "pso"],help="Whether the location of the output qubits can be different for the input location. Qubit locations can be optimized with pso.")
    # parser.add_argument("-f", "--full_reduce", dest="full_reduce", default=1, type=int, choices=[0,1], help="Full reduce")
    parser.add_argument("--population",nargs="+",default=30,type=int,help="The population size for the genetic algorithm.")
    parser.add_argument("--iterations",nargs="+",default=15,type=int,help="The number of iterations for the genetic algorithm.")
    parser.add_argument("--crossover_prob",nargs="+",default=0.8,type=restricted_float,help="The crossover probability for the genetic algorithm. Must be between 0.0 and 1.0.")
    parser.add_argument("--mutation_prob",nargs="+",default=0.2,type=restricted_float,help="The mutation probability for the genetic algorithm. Must be between 0.0 and 1.0.")
    # PSO args
    parser.add_argument("--swarm_size",nargs="+",default=15,type=int,help="The swarm size for the particle swarm optimizer.")
    parser.add_argument("--steps",nargs="+",default=10,type=int,help="The number of steps/iterations for the particle swarm optimizer.")
    parser.add_argument("--sco",nargs="+",default=0.4,type=restricted_float,help="The crossover percentage with the best particle in the swarm for the particle swarm optimizer. Must be between 0.0 and 1.0.")
    parser.add_argument("--pco",nargs="+",default=0.3,type=restricted_float,help="The crossover percentage with the personal best of a particle for the particle swarm optimizer. Must be between 0.0 and 1.0.")
    parser.add_argument("--mutation_perc",nargs="+",default=0.2,type=restricted_float,help="The mutation percentage of a particle for the particle swarm optimizer. Must be between 0.0 and 1.0.")
    # parser.add_argument("--perm", default="both", choices=["row", "col", "both"], help="Whether to find a single optimal permutation that permutes the rows, columns or both with the genetic algorithm.")
    parser.add_argument("--destination",help="Destination file or folder where the compiled circuit should be stored. Otherwise the source folder is used.")
    parser.add_argument("--metrics_csv",default=None,help="The location to store compiling metrics as csv, if not given, the metrics are not calculated. Only used when the source is a folder")
    parser.add_argument("--n_compile",default=1,type=int,help="How often to run the Quilc compiler, since it is not deterministic.")
    parser.add_argument("--subfolder", default=None,type=str,nargs="+",help="Possible subfolders from the main QASM source to compile from. Less typing when source folders are in the same folder. Can also be used for subfiles.")

    # args = parser.parse_args(args)
    args, unknown = parser.parse_known_args(args)
    if args.metrics_csv is not None and os.path.exists(args.metrics_csv):
        delete_csv = None
        text = input(
            "The given metrics file [%s] already exists. Do you want to overwrite it? (Otherwise it is appended) [y|n]"
            % args.metrics_csv
        )
        if text.lower() in ["y", "yes"]:
            delete_csv = True
        elif text.lower() in ["n", "no"]:
            delete_csv = False
        while delete_csv is None:
            text = input("Please answer yes or no.")
            if text.lower() in ["y", "yes"]:
                delete_csv = True
            elif text.lower() in ["n", "no"]:
                delete_csv = False
        if delete_csv:
            os.remove(args.metrics_csv)

    sources = make_into_list(args.QASM_source)
    if args.subfolder is not None:
        sources = [
            os.path.join(source, subfolder)
            for source in sources
            for subfolder in args.subfolder
            if os.path.isdir(source)
        ]
        # Remove non existing paths

    sources = [
        source
        for source in sources
        if os.path.exists(source)
        or print("Warning, skipping non-existing source:", source)
    ]
    sources = [s for s in sources if not s.endswith("py")]
    # print(sources)
    if "all" in args.mode:
        mode = elim_modes + [CompileMode.QUIL_COMPILER]
    else:
        mode = args.mode

    all_circuits = []
    for source in sources:
        print("Mapping qasm files in path:", source)
        circuits = batch_route_circuits(
            source,
            mode,
            args.architecture,
            n_qubits=args.qubits,
            populations=args.population,
            iterations=args.iterations,
            crossover_probs=args.crossover_prob,
            mutation_probs=args.mutation_prob,
            dest_folder=args.destination,
            metrics_file=args.metrics_csv,
            n_compile=args.n_compile,
            allow_output_perm=args.output_perm,
            swarm_size=args.swarm_size,
            n_steps=args.steps,
            s_crossover=args.sco,
            p_crossover=args.pco,
            pso_mutation=args.mutation_perc,
        )
        all_circuits.extend(circuits)


if __name__ == "__main__":
    print("Please call this as python -m pyzx router ...")



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import sys

from ..circuit import Circuit, determine_file_type
from .. import simplify
from .. import extract
from .. import optimize

description="""End-to-end circuit optimizer

For simple optimisation of a circuit run as
    python -m pyzx opt circuit.extension

This puts an optimised version of the circuit in the same directory and of the same file type.

If we want to specify the output location and type we can run
    python -m pyzx opt -d outputfile.qc -t qc inputfile.qasm
"""

import argparse
parser = argparse.ArgumentParser(prog="pyzx opt", description=description, formatter_class=argparse.RawTextHelpFormatter)
parser.add_argument('source',type=str,help='source circuit')
parser.add_argument('-d',type=str,help='destination for output file', dest='dest',default='')
parser.add_argument('-t',type=str,default='match', dest='outformat',
    help='Specify the output format (qasm, qc, quipper). By default matches the input')
parser.add_argument('-v',default=False, action='store_true', dest='verbose',
    help='Output verbose information')
parser.add_argument('-g',type=str,default='full', dest='simp', 
    help='ZX-simplifier to use. Options are full (default), cliff, or tele')
parser.add_argument('-p',default=False, action='store_true', dest='phasepoly',
    help='Whether to also run the phase-polynomial optimizer (default is false)')

def main(args):
    options = parser.parse_args(args)
    if not os.path.exists(options.source):
        print("File {} does not exist".format(options.source))
        return
    ctype = determine_file_type(options.source)
    if options.outformat == 'match':
        dtype = ctype
    elif options.outformat not in ('qasm', 'qc', 'quipper'):
        print("Unsupported circuit type {}. Please use qasm, qc or quipper".format(options.outformat))
        return
    else:
        dtype = options.outformat
    if not options.dest:
        base = os.path.splitext(options.source)[0]
        dest = base + "." + dtype
    else:
        dest = options.dest

    c = Circuit.load(options.source)
    if options.verbose:
        print("Starting circuit:")
        print(c.to_basic_gates().stats())
    g = c.to_graph()
    if options.verbose: print("Running simplification algorithm...")
    if options.simp == 'tele':
        g = simplify.teleport_reduce(g,quiet=(not options.verbose))
        c2 = Circuit.from_graph(g)
        c2 = c2.split_phase_gates()
    else:
        if options.simp == 'full':
            simplify.full_reduce(g,quiet=(not options.verbose))
        if options.simp == 'cliff':
            simplify.clifford_simp(g,quiet=(not options.verbose))
        if options.verbose: print("Extracting circuit...")
        c2 = extract.extract_circuit(g)
    if options.verbose: print("Optimizing...")
    if options.phasepoly:
        c3 = optimize.full_optimize(c2.to_basic_gates())
    else:
        c3 = optimize.basic_optimization(c2.to_basic_gates())
    c3 = c3.to_basic_gates()
    c3 = c3.split_phase_gates()
    if options.verbose: print(c3.stats())
    print("Writing output to {}".format(os.path.abspath(dest)))
    if dtype == 'qc': output = c3.to_qc()
    if dtype == 'qasm': output = c3.to_qasm()
    if dtype == 'quipper': output = c3.to_quipper()
    f = open(dest, 'w')
    f.write(output)
    f.close()



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
from collections import Counter
from typing import Any, Callable, Generic, Optional, List, Dict, Tuple

from ..utils import VertexType, EdgeType, FractionLike, FloatInt, phase_to_s
from .base import BaseGraph, VT, ET
from .graph_s import GraphS
from .jsonparser import string_to_phase

class GraphDiff(Generic[VT, ET]):
    removed_verts: List[VT]
    new_verts: List[VT]
    removed_edges: List[ET]
    new_edges: List[Tuple[Tuple[VT,VT],EdgeType]]
    changed_vertex_types: Dict[VT,VertexType]
    changed_edge_types: Dict[ET, EdgeType]
    changed_phases: Dict[VT, FractionLike]
    changed_pos: Dict[VT, Tuple[FloatInt,FloatInt]]
    changed_vdata: Dict[VT, Any]
    changed_edata: Dict[ET, Any]
    variable_types: Dict[str,bool]

    def __init__(self, g1: BaseGraph[VT,ET], g2: BaseGraph[VT,ET]) -> None:
        self.calculate_diff(g1,g2)

    def calculate_diff(self, g1: BaseGraph[VT,ET], g2: BaseGraph[VT,ET]) -> None:
        self.changed_vertex_types = {}
        self.changed_edge_types = {}
        self.changed_phases = {}
        self.changed_pos = {}
        self.changed_vdata = {}
        self.changed_edata = {}
        self.variable_types = g1.variable_types.copy()
        self.variable_types.update(g2.variable_types)

        old_verts = g1.vertex_set()
        new_verts = g2.vertex_set()
        self.removed_verts = list(old_verts - new_verts)
        self.new_verts = list(new_verts - old_verts)
        old_edges = g1.edge_set()
        new_edges = g2.edge_set()
        self.new_edges = []
        self.removed_edges = []

        for e in Counter(new_edges - old_edges).elements():
            self.new_edges.append((g2.edge_st(e), g2.edge_type(e)))

        for e in Counter(old_edges - new_edges).elements():
            self.removed_edges.append(e)

        for v in new_verts:
            if v in old_verts:
                if g1.type(v) != g2.type(v):
                    self.changed_vertex_types[v] = g2.type(v)
                if g1.phase(v) != g2.phase(v):
                    self.changed_phases[v] = g2.phase(v)
                d1 = g1.vdata_dict(v)
                d2 = g2.vdata_dict(v)
                if d1 != d2:
                    self.changed_vdata[v] = d2
                pos1 = g1.qubit(v), g1.row(v)
                pos2 = g2.qubit(v), g2.row(v)
                if pos1 != pos2:
                    self.changed_pos[v] = pos2
            else: # It is a new vertex
                if g2.type(v) != VertexType.Z: # We are taking the Z type to be the default
                    self.changed_vertex_types[v] = g2.type(v)
                if g2.phase(v) != 0:
                    self.changed_phases[v] = g2.phase(v)
                d2 = g2.vdata_dict(v)
                if d2:
                    self.changed_vdata[v] = d2
                pos2 = g2.qubit(v), g2.row(v)
                self.changed_pos[v] = pos2

        for e in new_edges:
            if e in old_edges:
                if g1.edge_type(e) != g2.edge_type(e):
                    self.changed_edge_types[e] = g2.edge_type(e)
                d1 = g1.edata_dict(e)
                d2 = g2.edata_dict(e)
                if d1 != d2:
                    self.changed_edata[e] = d2
            else:
                if g2.edge_type(e) != EdgeType.HADAMARD: # We take Hadamard edges to be the default
                    self.changed_edge_types[e] = g2.edge_type(e)
                d2 = g2.edata_dict(e)
                if d2:
                    self.changed_edata[e] = d2

    def apply_diff(self,g: BaseGraph[VT,ET]) -> BaseGraph[VT,ET]:
        g = copy.deepcopy(g)
        g.remove_edges(self.removed_edges)
        g.remove_vertices(self.removed_verts)
        for v in self.new_verts:
            g.add_vertex_indexed(v)
            g.set_position(v,*self.changed_pos[v])
            if v in self.changed_vertex_types:
                g.set_type(v,self.changed_vertex_types[v])
            else:
                g.set_type(v,VertexType.Z)
            if v in self.changed_phases:
                g.set_phase(v,self.changed_phases[v])
            if v in self.changed_vdata:
                g.set_vdata_dict(v, self.changed_vdata[v])
        for st, ty in self.new_edges:
            g.add_edge(st,ty)
        for e in self.changed_edata:
            for k, val in self.changed_edata[e].items():
                g.set_edata(e, k, val)

        for v in self.changed_pos:
            if v in self.new_verts: continue
            g.set_position(v,*self.changed_pos[v])

        for v in self.changed_vertex_types:
            if v in self.new_verts: continue
            g.set_type(v,self.changed_vertex_types[v])

        for v in self.changed_phases:
            if v in self.new_verts: continue
            g.set_phase(v,self.changed_phases[v])

        for v in self.changed_vdata:
            if v in self.new_verts: continue
            g.set_vdata_dict(v, self.changed_vdata[v])

        for e in self.changed_edge_types:
            g.set_edge_type(e,self.changed_edge_types[e])

        return g

    def to_dict(self) -> Dict[str, Any]:
        changed_edge_types_str_dict = {}
        for key, value in self.changed_edge_types.items():
            changed_edge_types_str_dict[f"{key[0]},{key[1]}"] = value # type: ignore
        changed_edata_str_dict = {}
        for key, value in self.changed_edata.items():
            changed_edata_str_dict[f"{key[0]},{key[1]}"] = value # type: ignore
        changed_phases_str = {k: phase_to_s(v) for k, v in self.changed_phases.items()}
        return {
            "removed_verts": self.removed_verts,
            "new_verts": self.new_verts,
            "removed_edges": self.removed_edges,
            "new_edges": self.new_edges,
            "changed_vertex_types": self.changed_vertex_types,
            "changed_edge_types": changed_edge_types_str_dict,
            "changed_phases": changed_phases_str,
            "changed_pos": self.changed_pos,
            "changed_vdata": self.changed_vdata,
            "changed_edata": changed_edata_str_dict,
            "variable_types": self.variable_types
        }

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    @staticmethod
    def from_json(json_str: str) -> "GraphDiff":
        d = json.loads(json_str)
        gd = GraphDiff(GraphS(),GraphS())
        gd.removed_verts = d["removed_verts"]
        gd.new_verts = d["new_verts"]
        gd.removed_edges = list(map(tuple, d["removed_edges"])) # type: ignore
        gd.new_edges = list(map(tuple, d["new_edges"])) # type: ignore
        gd.changed_vertex_types = map_dict_keys(d["changed_vertex_types"], int)
        gd.changed_edge_types = map_dict_keys(d["changed_edge_types"], lambda x: tuple(map(int, x.split(","))))
        gd.changed_phases = {int(k): string_to_phase(v,gd) for k, v in d["changed_phases"].items()}
        gd.changed_pos = map_dict_keys(d["changed_pos"], int)
        gd.changed_vdata = map_dict_keys(d["changed_vdata"], int)
        if "changed_edata" in d:
            gd.changed_edata = map_dict_keys(d["changed_edata"], lambda x: tuple(map(int, x.split(","))))
        else:
            gd.changed_edata = {}
        return gd

def map_dict_keys(d: Dict[str, Any], f: Callable[[str], Any]) -> Dict[Any, Any]:
    return {f(k): v for k, v in d.items()}



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import graph_tool.all as gt

from .base import BaseGraph

#DEPRECATED
class GraphGT(BaseGraph):
	backend = 'graph_tool'
	def __init__(self):
		raise Warning("Graph_tool is currently not fully supported.")
		BaseGraph.__init__(self)
		self.graph = gt.Graph(directed=False)
		self.graph.set_fast_edge_removal()
		self.graph.vp.type = self.graph.new_vertex_property('int')
		self.graph.ep.edata = self.graph.new_edge_property('object')

	def add_vertices(self, amount):
		self.graph.add_vertex(amount)

	def add_edges(self, edges):
		self.graph.add_edge_list(edges)

	def remove_vertices(self, vertices):
		for v in reversed(sorted(vertices)):
			self.graph.remove_vertex(v,fast=True)

	def remove_vertex(self, vertex):
		self.graph.remove_vertex(v)

	def remove_edges(self, edges):
		for e in edges:
			if isinstance(e,tuple):
				self.graph.remove_edge(self.graph.edge(e[0],e[1]))
			else:
				self.graph.remove_edge(e)

	def num_vertices(self):
		return self.graph.get_vertices().shape[0]

	def num_edges(self):
		return self.graph.get_edges().shape[0]

	def vertices(self):
		'''Iterator over all the vertices'''
		return self.graph.vertices()

	def verts_as_int(self, verts):
		return [int(v) for v in verts]

	def edges(self):
		'''Iterator over all the edges'''
		return self.graph.edges()

	def edges_as_int(self, edges):
		return [(self.graph.edge_index[e] if isinstance(e,gt.Edge) else e) for e in edges]

	def edge_set(self):
		#return set(self.graph.get_edges()[...,2])
		return set(self.graph.edges())

	def edge_st(self, edge):
		'''Returns a tuple of source/target of the given edge'''
		if isinstance(edge, gt.Edge):
			return (edge.source(), edge.target())
		e = gt.find_edge(self.graph,self.graph.edge_index,edge)[0]
		return (e.source(), e.target())

	def get_neighbors(self, vertex):
		'''Returns a tuple of source/target of the given edge'''
		return vertex.all_neighbors()

	def get_vertex_degree(self, vertex):
		'''Returns all neighboring vertices of the given vertex'''
		return vertex.in_degree() + vertex.out_degree()

	def get_incident_edges(self, vertex):
		#return [self.graph.edge_index[e] for e in vertex.all_edges()]
		return vertex.all_edges()

	def is_connected(self,v1,v2):
		return bool(self.graph.edge(v1, v2))

	def is_equal(self,v1,v2):
		'''Returns whether v1 and v2 represent the same vertex'''
		return int(v1) == int(v2)

	def get_type(self, vertex):
		return self.graph.vp.type[vertex]

	def set_type(self, vertex, t):
		self.graph.vp.type[vertex] = t

	def add_attribute(self, attrib_name, default=0):
		self.graph.vertex_properties[attrib_name] = self.graph.new_vertex_property('int')
		self.graph.vertex_properties[attrib_name].set_value(default)

	def clear_edata(self, edge):
		self.graph.ep.edata[edge] = None

	def edata_keys(self, edge):
		edata = self.graph.ep.edata[edge]
		return edata.keys() if edata else []

	def edata(self, edge, key, default=None):
		edata = self.graph.ep.edata[edge]
		if edata and key in edata:
			return edata[key]
		return default

	def set_edata(self, edge, key, val):
		edata = self.graph.ep.edata[edge]
		if edata:
			edata[key] = val
		else:
			self.graph.ep.edata[edge] = {key: val}



# PyZX - Python library for quantum circuit rewriting 
#        and optimisation using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
	import igraph as ig
except ImportError:
	print("python-igraph not available")
	ig = None

from .base import BaseGraph, VertexType, EdgeType

class GraphIG(BaseGraph):
	"""Implementation of :class:`~graph.base.BaseGraph` using ``python-igraph`` 
	as its backend"""
	backend = 'igraph'
	def __init__(self):
		raise Warning("Python-igraph is currently not fully supported.")
		BaseGraph.__init__(self)
		self.graph = ig.Graph(directed=False)
		self.graph.vs['_a'] = None
		self.graph.vs['_t'] = None
		self.graph.vs['_q'] = None
		self.graph.vs['_r'] = None
		self.graph.es['_t'] = None
		self.graph.es['_edata'] = None
		self._maxq = -1
		self._maxr = -1
		self.inputs = []
		self.outputs = []

	# since igraph uses consecutive indexing, vindex() == num_vertices()
	def vindex(self):
		return self.num_vertices()

	def depth(self): 
		self._maxr = max(self._rindex.values())
		return self._maxr
	def qubit_count(self): return self._maxq + 1

	def add_vertices(self, amount, vertex_data=None):
		self.graph.add_vertices(amount)
		n = self.num_vertices()
		return range(n - amount, n)

	def add_edges(self, edges, edge_data=None):
		self.graph.add_edges(edges)

	def remove_vertices(self, vertices):
		self.graph.delete_vertices(vertices)

	def remove_isolated_vertices(self):
		self.graph.vs.select(_degree=0).delete()

	def remove_edges(self, edges):
		self.graph.delete_edges(edges)

	def num_vertices(self):
		return len(self.graph.vs)

	def num_edges(self):
		return len(self.graph.es)

	def vertices(self):
		return range(len(self.graph.vs))

	def edges(self):
		return range(len(self.graph.es))

	def edge(self, s, t):
		return self.graph.es[s,t][0].index
	
	def edge_st(self, edge):
		edge = self.graph.es[edge]
		return (edge.source, edge.target)

	def edge_s(self, edge):
		edge = self.graph.es[edge]
		return edge.source

	def edge_t(self, edge):
		edge = self.graph.es[edge]
		return edge.target

	def neighbors(self, vertex):
		return self.graph.neighbors(vertex)

	def vertex_degree(self, vertex):
		return self.graph.degree(vertex)

	def incident_edges(self, vertex):
		return self.graph.incident(vertex, mode=ig.ALL)

	def connected(self,v1,v2):
		return self.graph.are_connected(v1,v2)

	def edge_type(self, e):
		if e in self.graph.es:
			t = self.graph.es[e]['_t']
			return 1 if t == None else t
		else:
			return 0

	def set_edge_type(self, e, t):
		self.graph.es[e]['_t'] = t


	def type(self, v):
		t = self.graph.vs[v]['_t']
		return t if t is not None else VertexType.BOUNDARY

	def types(self):
		return self.graph.vs['_t']

	def set_type(self, v, t):
		self.graph.vs[v]['_t'] = t
		
	def phase(self, vertex):
		a = self.graph.vs[vertex]['_a']
		return a if a != None else 0

	def phases(self):
		return [a if a != None else 0 for a in self.graph.vs['_a']]

	def set_phase(self, vertex, phase):
		self.graph.vs[vertex]['_a'] = phase % 2

	def qubit(self, vertex):
		return self.graph.vs[vertex]['_q'] or -1
	def qubits(self):
		return self.graph.vs['_q']
	def set_qubit(self, vertex, q):
		if q > self._maxq: self._maxq = q
		self.graph.vs[vertex]['_q'] = q

	def row(self, vertex):
		return self.graph.vs[vertex]['_r'] or -1
	def rows(self):
		return self.graph.vs['_r']
	def set_row(self, vertex, r):
		if r > self._maxr: self._maxr = r
		self.graph.vs[vertex]['_r'] = r

	def set_vdata(self, v, key, val):
		self.graph.vs[v][key] = val

	def vdata_keys(self, v):
		return [a for a in self.graph.vertex_attributes() if a != '_a' and a != '_t' and a != '_q' and a != '_r']

	def vdata(self, v, key, default=None):
		try:
			val = self.graph.vs[v][key]
			if not val: val = default
		except KeyError:
			val = default
		return val

	def clear_edata(self, edge):
		self.graph.es[edge]['_edata'] = None
	def edata_keys(self, edge):
		edata = self.graph.es[edge]['_edata']
		return edata.keys() if edata else []
	def edata(self, edge, key, default=None):
		edata = self.graph.es[edge]['_edata']
		if edata and key in edata:
			return edata[key]
		return default
	def set_edata(self, edge, key, val):
		edata = self.graph.es[edge]['_edata']
		if edata:
			edata[key] = val
		else:
			self.graph.es[edge]['_edata'] = {key: val}



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""This module implements the Third Order Duplicate and Destroy (TODD) algorithm
from Luke E Heyfron and Earl T Campbell 2019 Quantum Sci. Technol. 4 015004
available at http://iopscience.iop.org/article/10.1088/2058-9565/aad604/meta

The main entry point is :func:`todd_simp`, which is used in :func:`pyzx.optimize.phase_block_optimize`."""

from __future__ import print_function

from fractions import Fraction
import subprocess
import tempfile
import time
import random
from typing import Optional, Dict, Tuple, List, Set, Iterable, cast

import numpy as np

from .circuit.gates import Gate, T, S, Z, ZPhase, CZ, CNOT, ParityPhase
from .utils import settings, EdgeType, VertexType, FractionLike
from .graph.base import BaseGraph, VT, ET, upair
from .linalg import Mat2, MatLike, Z2
from .extract import permutation_as_swaps, column_optimal_swap
from .parity_network import parity_network
from .symbolic import Poly

USE_REED_MULLER: bool = False


class ParityPolynomial(object):
    """Class used to represent phase polynomials in the standard
    ParityPhase view. For example: x1@x2 + 3x2 + 5x1@x2@x3"""
    terms: Dict[Tuple[int,...],int]
    def __init__(self,qubits:int, poly:Optional['ParityPolynomial']=None) -> None:
        self.qubits = qubits
        if poly is not None:
            self.terms = poly.terms.copy()
        else: self.terms = {}
    
    def copy(self) -> 'ParityPolynomial':
        return type(self)(self.qubits, self)
    
    def __str__(self) -> str:
        l = []
        for t in sorted(self.terms.keys()):
            val = self.terms[t]
            l.append("{!s}{}".format(val if val!=1 else "", "@".join("x{:d}".format(v) for v in sorted(list(t)))))
        return " + ".join(l)
    
    def __repr__(self):
        return str(self)
    
    def add_term(self, term: Iterable[int], value: int) -> None:
        term = tuple(sorted(term))
        if term in self.terms:
            self.terms[term] = (self.terms[term] + value) % 8
        else: self.terms[term] = value % 8
        if not self.terms[term]:
            del self.terms[term]
    
    def add_polynomial(self, poly: 'ParityPolynomial') -> None:
        for term, val in poly.terms.items():
            self.add_term(term, val)
    
    def __add__(self, other: 'ParityPolynomial') -> 'ParityPolynomial':
        p = self.copy()
        p.add_polynomial(other)
        return p
    
    def to_par_matrix(self) -> Mat2:
        """Converts the phase polynomial into a parity matrix."""
        cols = []
        for par, val in self.terms.items():
            col: List[Z2] = [1 if i in par else 0 for i in range(self.qubits)]
            for i in range(val): cols.append(col)
        return Mat2(cols).transpose()

class ParitySingle(object):
    """Class used for representing a single parity expression
    like x1@x2@x4"""
    def __init__(self,startval: int) -> None:
        self.par: Set[int] = {startval}
    
    def __str__(self) -> str:
        return "@".join("x{:d}".format(i) for i in sorted(self.par))
    
    def __repr__(self) -> str:
        return str(self)
    
    def add_par(self, other: 'ParitySingle') -> None:
        self.par.symmetric_difference_update(other.par)


class MultiLinearPoly(object):
    """Class for representing phase polynomials in the multilinear formalism.
    For example: x1 + x2 + 2x1x2 + 4x1x2x3"""
    def __init__(self) -> None:
        self.l: Dict[int,int]            = {}
        self.q: Dict[Tuple[int,int],int] = {}
        self.c: Set[Tuple[int,int,int]]  = set()
    
    def add_parity(self, par:Iterable[int], subtract:bool=False) -> None:
        p = []
        mult = -1 if subtract else 1
        for i,v in enumerate(par):
            if v: p.append(i)
        for a in range(len(p)):
            v1 = p[a]
            if v1 not in self.l: self.l[v1] = mult
            else: self.l[v1] = (self.l[v1] + mult) % 8
            
            for b in range(a+1, len(p)):
                v2 = p[b]
                if (v1,v2) not in self.q: self.q[(v1,v2)] = 1 if subtract else 3
                else: self.q[(v1,v2)] = (self.q[(v1,v2)] - mult) % 4
                    
                for c in range(b+1, len(p)):
                    v3 = p[c]
                    if (v1,v2,v3) not in self.c: self.c.add((v1,v2,v3))
                    else: self.c.remove((v1,v2,v3))
    
    def add_par_matrix(self, a: Mat2, subtract:bool=False) -> None:
        for col in a.transpose().data:
            self.add_parity(col,subtract=subtract)
    
    def to_clifford(self) -> List[Gate]:
        """Returns the phase polynomial in terms of Clifford Z-rotations 
        and CZs. If the phase polyomial is not Clifford it raises an ValueError."""
        gates: List[Gate] = []
        for t, v in self.l.items():
            if v == 2:
                gates.append(S(t,adjoint=False))
            elif v == 4:
                gates.append(Z(t))
            elif v == 6:
                gates.append(S(t,adjoint=True))
            elif v != 0:
                raise ValueError("PhasePoly is not Clifford")
        for (t1,t2), v in self.q.items():
            if v == 2:
                gates.append(CZ(t1,t2))
            elif v != 0:
                raise ValueError("PhasePoly is not Clifford")
        if self.c:
            raise ValueError("PhasePoly is not Clifford")
        return gates


def par_matrix_to_gates(a: Mat2) -> List[Gate]:
    """Convert a parity matrix into T gates and ParityPhase gates."""
    gates: List[Gate] = []
    phase = Fraction(1,4)
    for col in a.transpose().data:
        targets = [i for i,v in enumerate(col) if v]
        if len(targets) == 1:
            gates.append(T(targets[0]))
        else:
            gates.append(ParityPhase(phase, *targets))
    return gates

def phase_gates_to_poly(gates: List[Gate], qubits: int) -> Tuple[ParityPolynomial,List[ParitySingle]]:
    """Convert a CNOT+T+CZ circuit into a phase polynomial representation
    using :class:`ParityPolynomial`."""
    phase_poly = ParityPolynomial(qubits)
    expression_polys = []
    for i in range(qubits):
        expression_polys.append(ParitySingle(i))

    for g in gates:
        if isinstance(g, ZPhase):
            par = expression_polys[g.target].par
            g_phase = g.phase
            if isinstance(g_phase, Poly):
                raise NotImplementedError("Symbolic phases not supported")
            phase_poly.add_term(par, int(float(g_phase*4)))
        elif isinstance(g, CZ):
            tgt, ctrl = g.target, g.control
            par1 = expression_polys[tgt].par
            par2 = expression_polys[ctrl].par
            phase_poly.add_term(par1, 2)
            phase_poly.add_term(par2, 2)
            phase_poly.add_term(par1.symmetric_difference(par2), 6)
        elif isinstance(g, CNOT):
            tgt, ctrl = g.target, g.control
            expression_polys[tgt].add_par(expression_polys[ctrl])
        else:
            raise TypeError("Unknown gate type {}".format(str(g)))
    
    return phase_poly, expression_polys



def xi(m: Mat2, z: List[Z2]) -> Mat2:
    """Constructs the \\chi matrix from the TOpt paper."""
    arr = np.asarray(m.data)
    rows = m.rows()
    data = []
    for alpha in range(rows):
        ra = arr[alpha]
        for beta in range(alpha+1, rows):
            rb = arr[beta]
            rab = ra*rb
            for gamma in range(beta+1, rows):
                rg = arr[gamma]
                if z[alpha]:
                    rbg = rb*rg
                    if not z[beta]:
                        if not z[gamma]:
                            data.append(rbg.tolist())
                            continue
                        data.append(((rbg+rab)%2).tolist())
                        continue
                    elif not z[gamma]:
                        rag = ra*rg
                        data.append(((rbg+rag)%2).tolist())
                        continue
                    else: #z[alpha], z[beta] and z[gamma] are all true
                        rag = ra*rg
                        data.append(((rab+rag+rbg)%2).tolist())
                        continue
                elif z[beta]:
                    rag = ra*rg
                    if z[gamma]:
                        data.append(((rab+rag)%2).tolist())
                        continue
                    data.append(rag.tolist())
                    continue
                elif z[gamma]:
                    data.append(rab.tolist())
                    continue
    for r in m.data: data.append(r.copy())            
    return Mat2(data)


def find_todd_match(m: Mat2) -> Tuple[int,int, Optional[List[Z2]],Optional[List[Z2]]]:
    """Tries to find a match for the TODD algorithm given a parity matrix."""
    rows = m.rows()
    cols = m.cols()
    for a in range(cols):
        for b in range(a+1, cols):
            z: List[Z2] = [0]*rows
            for i in range(rows):
                r = m.data[i]
                if r[a]:
                    if not r[b]:
                        z[i] = 1
                else:
                    if r[b]:
                        z[i] = 1
            bigm = xi(m, z)
            #print(bigm, '.')
            options = bigm.nullspace(should_copy=False)
            #print(bigm)
            for y in options:
                if y[a] + y[b] == 1: return a,b,z,y

    return -1,-1,None,None


def remove_trivial_cols(m: Mat2) -> int:
    """Remove duplicate and zero columns in parity matrix.
    NOTE: the transpose of the matrix should be supplied
    so that the columns are actually the rows."""
    while True:
        newcols = m.rows()
        for a in range(newcols):
            if not any(m.data[a]):
                m.data.pop(a)
                break
            should_break = False
            for b in range(a+1, newcols):
                if m.data[a] == m.data[b]:
                    m.data.pop(b)
                    m.data.pop(a)
                    should_break = True
                    break
            if should_break: break
        else: # Didn't break out of for-loop so didn't find any match
            break
    return newcols

def do_todd_single(m: Mat2) -> Tuple[Mat2,int]:
    """Find a single TODD match and apply it to the matrix."""
    startcols = m.cols()
    a,b,z,y = find_todd_match(m)
    if z is None: return m, 0
    assert y is not None
    m = m.transpose()
    #odd_y = sum(y) % 2
    for i,c in enumerate(m.data):
        if not y[i]: continue
        for j in range(len(c)):
            if z[j]: c[j] = 0 if c[j] else 1
    if sum(y) % 2 == 1:
        m.data.append(z)
    m.data.pop(b)
    m.data.pop(a)
    
    newcols = remove_trivial_cols(m)
                
    return m.transpose(), startcols - newcols

def todd_iter(m: Mat2, quiet:bool=True) -> Mat2:
    """Keep finding TODD matches until nothing is found anymore.
    If ``zx.settings.topt_command`` is set it uses the TOpt implementation of TODD."""
    m = m.transpose()
    remove_trivial_cols(m)
    random.shuffle(m.data) # Randomly shuffle the columns
    m = m.transpose()
    if not m.cols() or not m.rows():
        return m
    if settings.topt_command is not None:
        return call_topt(m, quiet=quiet)
    while True:
        m, reduced = do_todd_single(m)
        if reduced == 0:
            if not quiet: print()
            return m
        if not quiet: print(reduced, end='.')

def call_topt(m: Mat2, quiet:bool=True) -> Mat2:
    """Calls and parses the output of the TOpt implementation of TODD."""
    assert settings.topt_command is not None
    if not quiet:
        print("TOpt: ", end="")
    t_start = m.cols()
    s = "\n".join(" ".join(str(i) for i in r) for r in m.data)
    with tempfile.NamedTemporaryFile(suffix='.gsm') as f:
        f.write(s.encode('ascii'))
        f.flush()
        time.sleep(0.01)
        if settings.topt_command[0].find('wsl') != -1:
            fname = "/mnt/c"+f.name.replace("\\", "/")[2:]
        else: fname = f.name
        if USE_REED_MULLER:
            output = subprocess.check_output([*settings.topt_command, "gsm",fname, "-a", "rm"])
        else:
            output = subprocess.check_output([*settings.topt_command, "gsm",fname])
        out = output.decode()
        #print(out)
    rows = out[out.find("Output gate"):out.find("Successful")].strip().splitlines()[2:]
    i = out.find("Total time")
    t = out[i+10: out.find("s",i)]
    if not quiet:
        print(t)
    data: MatLike = []
    try:
        for row in rows:
            data.append([cast(Z2, i) for i in row])
    except ValueError:
        print(out)
        print(rows)
        raise
    m2 = Mat2(data)
    if USE_REED_MULLER:
        m = m2.transpose()
        remove_trivial_cols(m)
        m2 = m.transpose()
    t_end = m2.cols()
    if t_end < t_start:
        if not quiet: print("Found reduction: ", t_start - t_end)
        # print("Start:")
        # print(m)
        # print("End:")
        # print(m2)
        #print(out)
    return m2


def todd_simp(gates: List[Gate], qubits: int, quiet:bool=True) -> Tuple[List[Gate],Dict[int,int]]:
    """Run the TODD algorithm on a CNOT+CZ+T set of gates and 
    apply the necessary Clifford corrections. Uses the 
    CNOT parity algorithm from https://arxiv.org/pdf/1712.01859.pdf
    to synthesize the necessary parities."""
    phase_poly, parity_polys = phase_gates_to_poly(gates, qubits)
    #print(phase_poly)
    #print(parity_polys)
    m = phase_poly.to_par_matrix()
    m2 = todd_iter(m,quiet=quiet)

    newgates: List[Gate] = []
    parities = []
    for col in m2.transpose().data:
        if sum(col) == 1:
            newgates.append(T(next(i for i in range(qubits) if col[i])))
        else:
            parities.append(col)

    p = MultiLinearPoly()
    p.add_par_matrix(m,False)
    p.add_par_matrix(m2,True)
    newgates.extend(p.to_clifford())

    cnots = parity_network(qubits, parities)
    m = Mat2.id(qubits)
    for cnot in cnots:
        m.row_add(cnot.control, cnot.target)
    data = []
    for q in parity_polys:
        l = [int(i in q.par) for i in range(qubits)]
        data.append(l)
    target_matrix = Mat2(data) * m.inverse() # type: ignore
    #perm = column_optimal_swap(target_matrix.transpose())
    perm = {i:i for i in range(qubits)}
    swaps = permutation_as_swaps(perm)
    for a,b in swaps:
        target_matrix.row_swap(a,b)
    cnot_list = target_matrix.to_cnots(optimize=True)
    for gate in reversed(cnot_list):
        cnots.append(CNOT(gate.target,gate.control))

    m = Mat2.id(qubits)
    for i, cnot in enumerate(cnots):
        newgates.append(cnot)
        m.row_add(cnot.control, cnot.target)
        for par in parities:
            if par in m.data: # The parity checks out, so put a phase here
                newgates.append(T(m.data.index(par)))
                parities.remove(par)
                break

    if parities:
        raise ValueError("Still phases left on the stack")

    return newgates, {v:k for k,v in perm.items()}


def todd_on_graph(g: BaseGraph[VT,ET]) -> None:
    """Runs the TODD algorithm on a graph. The variables are determined
    by looking at which vertices have phase gadgets attached to them.
    Note that this produces graphs that can only be transformed into circuits
    using ancilla qubits."""
    gadgets = {}
    t_nodes = []
    inputs = g.inputs()
    outputs = g.outputs()
    for v in g.vertices():
        if v not in inputs and v not in outputs and len(list(g.neighbors(v)))==1:
            if g.phase(v) != 0 and g.phase(v).denominator != 4: continue
            n = list(g.neighbors(v))[0]
            tgts = frozenset(set(g.neighbors(n)).difference({v}))
            gadgets[tgts] = (n,v)
        if g.phase(v) != 0 and g.phase(v).denominator == 4:
            t_nodes.append(v)
    
    if not gadgets:
        print("No phase gadgets found")
        return
    variables: Set[VT] = set()
    for par in gadgets.keys():
        variables.update(par)
    
    for v in variables:
        if v in t_nodes:
            gadgets[frozenset({v})] = (v,v)
    
    targets = list(variables)
    nt = len(targets)

    cols = []
    for par, (_,v) in gadgets.items():
        col: List[Z2] = [0]*nt
        for t in par:
            col[targets.index(t)] = 1
        phase = g.phase(v)
        for i in range(phase.numerator): cols.append(col)
    parmatrix = Mat2(cols).transpose()
    m2 = todd_iter(parmatrix)
    
    newgadgets = []
    phases: Dict[VT,FractionLike] = dict()
    for col in m2.transpose().data:
        if sum(col) == 1:
            i = next(i for i,a in enumerate(col) if a)
            v = targets[i]
            if v in t_nodes:
                phases[v] = Fraction(1,4)
            else:
                phases[v] = g.phase(v) + Fraction(1,4)
        else:
            newgadgets.append(frozenset([targets[i] for i,a in enumerate(col) if a]))
    
    p = MultiLinearPoly()
    p.add_par_matrix(parmatrix,False)
    p.add_par_matrix(m2,True)
    correction = p.to_clifford()
    add_czs: Dict[Tuple[VT,VT], List[int]] = {}
    for clif in correction:
        if isinstance(clif, ZPhase):
            v = targets[clif.target]
            if v in phases:
                phases[v] += clif.phase
            else:
                if v in t_nodes:
                    phases[v] = clif.phase
                else:
                    phases[v] = g.phase(v) + clif.phase
        elif clif.name == 'CZ':
            v1,v2 = targets[clif.control], targets[clif.target] # type: ignore
            add_czs[upair(v1,v2)] = [0,1]
        else:
            raise ValueError("Unknown clifford correction:", str(clif))
    
    for v in targets:
        if v in phases:
            g.set_phase(v, phases[v])
        else:
            if v in t_nodes:
                g.set_phase(v, 0)
    g.add_edge_table(add_czs)
    
    rs = g.rows()
    positions = set()
    for gadget, (n,v) in gadgets.items():
        if len(gadget) == 1: continue # T-node
        if gadget in newgadgets:
            positions.add(rs[v])
            g.set_phase(v, Fraction(1,4))
            newgadgets.remove(gadget)
        else:
            g.remove_vertices((n,v))
    
    edges: List[Tuple[VT,VT]] = []
    for par in newgadgets:
        pos = sum(rs[t] for t in par)/len(par) + 0.5
        while pos in positions: pos += 0.5
        n = g.add_vertex(VertexType.Z, -1, pos)
        v = g.add_vertex(VertexType.Z, -2, pos, phase=Fraction(1,4))
        edges.append((n,v))
        positions.add(pos)
        for t in par: edges.append((n,t))
    g.add_edges(edges, EdgeType.HADAMARD)



# PyZX - Python library for quantum circuit rewriting
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import List, Union, Optional, Iterator, Dict

import numpy as np

from .gates import (Gate, gate_types, NOT, Y, Z, HAD, XPhase, YPhase, ZPhase, U2, U3, S, T, SX, SWAP, RXX, RZZ, CNOT,
                    CY, CZ, CHAD, CSX, XCX, CRX, CRY, CRZ, CPhase, CU3, CU, CSWAP, Tofolli, CCZ, ParityPhase, FSim,
                    Measurement)

from ..graph.base import BaseGraph
from ..utils import EdgeType

CircuitLike = Union['Circuit', Gate]

# Note that many of the method of Circuit contain inline imports. These are
# there to prevent circular imports.

__all__ = ['Circuit', 'id']

class Circuit(object):
    """Class for representing quantum circuits.

    This class is mostly just a wrapper for a list of gates with methods for converting
    between different representations of a quantum circuit.

    The methods in this class that convert a specification of a circuit into an instance of this class,
    generally do not check whether the specification is well-defined. If a bad input is given,
    the behaviour is undefined."""
    def __init__(self, qubit_amount: int, name: str = '', bit_amount: Optional[int] = None) -> None:
        self.qubits: int        = qubit_amount
        self.bits: int = 0 if bit_amount is None else bit_amount
        self.gates:  List[Gate] = []
        self.name:   str        = name

    ### BASIC FUNCTIONALITY


    def __str__(self) -> str:
        return "Circuit({!s} qubits, {!s} bits, {!s} gates)".format(self.qubits,self.bits,len(self.gates))

    def __repr__(self) -> str:
        return str(self)

    def copy(self) -> 'Circuit':
        c = Circuit(self.qubits, self.name, self.bits)
        c.gates = [g.copy() for g in self.gates]
        return c

    def adjoint(self) -> 'Circuit':
        c = Circuit(self.qubits, self.name + 'Adjoint', self.bits)
        for g in reversed(self.gates):
            c.gates.append(g.to_adjoint())
        return c


    def verify_equality(self, other: 'Circuit', up_to_swaps: bool = False, up_to_global_phase: bool = True) -> bool:
        """Composes the other circuit with the adjoint of this circuit, and tries to reduce
        it to the identity using :func:`simplify.full_reduce``. If successful returns True,
        if not returns None.

        Note:
            A successful reduction to the identity is strong evidence that the two
            circuits are equal, if this function is not able to reduce the graph to the identity
            this does not prove anything.

        Args:
            other: the circuit to compare equality to.
            up_to_swaps: if set to True, only checks equality up to a permutation of the qubits.

        """
        if self.bits or other.bits:
            # TODO once full_gnd_reduce is merged
            raise NotImplementedError("The equality verification does not support hybrid circuits.")

        from ..simplify import full_reduce
        c = self.adjoint()
        c.add_circuit(other)
        g = c.to_graph()
        full_reduce(g)
        if (g.num_vertices() == self.qubits*2 and
                all(g.edge_type(e) == EdgeType.SIMPLE for e in g.edges())):
            return ((up_to_swaps or all(g.connected(v,w) for v,w in zip(g.inputs(),g.outputs()))) and
                    (up_to_global_phase or g.scalar.phase == 0))
        else:
            return False

    def add_gate(self, gate: Union[Gate,str], *args, **kwargs) -> None:
        """Adds a gate to the circuit. ``gate`` can either be
        an instance of a :class:`Gate`, or it can be the name of a gate,
        in which case additional arguments should be given.

        Example::

            circuit.add_gate("CNOT", 1, 4) # adds a CNOT gate with control 1 and target 4
            circuit.add_gate("ZPhase", 2, phase=Fraction(3,4)) # Adds a ZPhase gate on qubit 2 with phase 3/4
        """
        if isinstance(gate, str):
            gate_class = gate_types[gate]
            gate = gate_class(*args, **kwargs)
        self.gates.append(gate)

    def prepend_gate(self, gate, *args, **kwargs):
        """The same as add_gate, but adds the gate to the start of the circuit, not the end.
        """
        if isinstance(gate, str):
            gate_class = gates.gate_types[gate]
            gate = gate_class(*args, **kwargs)
        self.gates.insert(0, gate)

    def add_gates(self, gates: str, qubit: int) -> None:
        """Adds a series of single qubit gates on the same qubit.
        ``gates`` should be a space-separated string of gatenames.

        Example::

            circuit.add_gates("S T H T H", 1)
        """
        for g in gates.split(" "):
            self.add_gate(g, qubit)

    def add_circuit(self, circ: 'Circuit', mask: Optional[List[int]]=None, bit_mask: Optional[List[int]]=None) -> None:
        """Adds the gate of another circuit to this one. If ``mask`` is not given,
        then they must have the same amount of qubits and they are mapped one-to-one.
        If mask is given then it must be a list specifying to which qubits the qubits
        in the given circuit correspond. Similarly, if ``bit_mask`` is not given,
        then they must have the same amount of bits.

        Example::

            c1 = Circuit(qubit_amount=4)
            c2 = Circuit(qubit_amount=2)
            c2.add_gate("CNOT",0,1)
            c1.add_circuit(c2, mask=[0,3]) # Now c1 has a CNOT from the first to the last qubit

        If the circuits have the same amount of qubits then it can also be called as an operator::

            c1 = Circuit(2)
            c2 = Circuit(2)
            c1 += c2

        """
        if mask is None and bit_mask is None:
            if self.qubits != circ.qubits:
                raise TypeError("Amount of qubits do not match")
            if self.bits != circ.bits:
                raise TypeError("Amount of bits do not match")
            self.gates.extend([g.copy() for g in circ.gates])
            return
        if mask is None:
            mask = list(range(self.qubits))
        if bit_mask is None:
            bit_mask = list(range(self.bits))
        if len(mask) != circ.qubits:
            raise TypeError("Mask size does not match qubits")
        if len(bit_mask) != circ.bits:
            raise TypeError("Bit mask size does not match bits")
        for gate in circ.gates:
            g = gate.reposition(mask, bit_mask)
            self.add_gate(g)

    def tensor(self, other: CircuitLike) -> 'Circuit':
        """Takes the tensor product of two Circuits. Places the second one below the first.
        Can also be done as an operator: `circuit1 @ circuit2`."""
        if isinstance(other,Gate):
            c2 = Circuit(other._max_target()+1)
            c2.add_gate(other)
            other = c2
        if not isinstance(other,Circuit):
            raise Exception("Cannot tensor type", type(other), "to Circuit")
        c = Circuit(self.qubits + other.qubits)
        c.gates = [g.copy() for g in self.gates]
        mask = [i+self.qubits for i in range(other.qubits)]
        bit_mask = [i+self.bits for i in range(other.bits)]
        c.gates.extend([g.reposition(mask, bit_mask) for g in other.gates])
        return c

    def to_basic_gates(self) -> 'Circuit':
        """Returns a new circuit with every gate expanded in terms of X/Z phases, Hadamards
        and the 2-qubit gates CNOT, CZ, CX."""
        c = Circuit(self.qubits, name=self.name, bit_amount=self.bits)
        for g in self.gates:
            c.gates.extend(g.to_basic_gates())
        return c

    def split_phase_gates(self) -> 'Circuit':
        c = Circuit(self.qubits, name=self.name)
        for g in self.gates:
            if isinstance(g, (ZPhase, XPhase)):
                c.gates.extend(g.split_phases())
            else:
                c.add_gate(g)
        return c

    ### OPERATORS

    def __iadd__(self, other: CircuitLike) -> 'Circuit':
        if isinstance(other, Circuit):
            self.add_circuit(other)
            if other.qubits > self.qubits:
                self.qubits = other.qubits
        elif isinstance(other, Gate):
            self.add_gate(other)
            if other._max_target() + 1 > self.qubits:
                self.qubits = other._max_target() + 1
        else:
            raise Exception("Cannot add object of type", type(other), "to Circuit")
        return self

    def __add__(self, other: CircuitLike) -> 'Circuit':
        c = self.copy()
        c += other
        return c

    def __len__(self) -> int:
        return len(self.gates)

    def __iter__(self) -> Iterator[Gate]:
        return iter(self.gates)

    def __matmul__(self, other: CircuitLike) -> 'Circuit':
        return self.tensor(other)


    ### MATRIX EMULATION (FOR E.G. Mat2.gauss)

    def row_add(self, q0: int, q1: int):
        self.add_gate("CNOT", q0, q1)

    def col_add(self, q0: int, q1: int):
        self.prepend_gate("CNOT", q1, q0)


    ### CONVERSION METHODS


    @staticmethod
    def from_graph(g:BaseGraph, split_phases:bool=True) -> 'Circuit':
        """Produces a :class:`Circuit` containing the gates of the given ZX-graph.
        If the ZX-graph is not circuit-like then the behaviour of this function
        is undefined.
        ``split_phases`` governs whether nodes with phases should be split into
        Z,S, and T gates or if generic ZPhase/XPhase gates should be used."""
        from .graphparser import graph_to_circuit
        return graph_to_circuit(g, split_phases=split_phases)

    def to_graph(self, zh:bool=False, compress_rows:bool=True, backend:Optional[str]=None) -> BaseGraph:
        """Turns the circuit into a ZX-Graph.
        If ``compress_rows`` is set, it tries to put single qubit gates on different qubits,
        on the same row."""
        from .graphparser import circuit_to_graph

        return circuit_to_graph(self if zh else self.to_basic_gates(),
            compress_rows, backend)

    def to_tensor(self, preserve_scalar:bool=True) -> np.ndarray:
        """Returns a numpy tensor describing the circuit."""
        return self.to_graph().to_tensor(preserve_scalar)
    def to_matrix(self, preserve_scalar=True) -> np.ndarray:
        """Returns a numpy matrix describing the circuit."""
        return self.to_graph().to_matrix(preserve_scalar)

    def to_emoji(self) -> str:
        """Converts circuit into a representation that can be copy-pasted
        into the ZX-calculus Discord server."""
        from .emojiparser import circuit_to_emoji
        return circuit_to_emoji(self)

    @staticmethod
    def load(circuitfile: str) -> 'Circuit':
        """Tries to detect the circuit description language from the filename and its contents,
        and then tries to load the file into a circuit."""
        from .quipperparser import quipper_center_block

        ext = determine_file_type(circuitfile)
        if ext == 'qc':
            return Circuit.from_qc_file(circuitfile)
        if ext == 'qasm':
            return Circuit.from_qasm_file(circuitfile)
        if ext == 'qsim':
            return Circuit.from_qsim_file(circuitfile)
        if ext == 'qgraph':
            raise TypeError(".qgraph files are not Circuits. Please load them as graphs using json_to_graph")
        if ext == 'quipper':
            return Circuit.from_quipper_file(circuitfile)
        raise TypeError("Couldn't determine filetype")

    @staticmethod
    def from_qc(s: str) -> 'Circuit':
        from .qcparser import parse_qc
        return parse_qc(s)

    @staticmethod
    def from_qc_file(fname: str) -> 'Circuit':
        """Produces a :class:`Circuit` based on a .qc description of a circuit.
        If a Toffoli gate with more than 2 controls is encountered, ancilla qubits are added.
        Currently up to 5 controls are supported."""
        from .qcparser import parse_qc
        with open(fname, 'r') as f:
            data = f.read()
        c = parse_qc(data)
        c.name = os.path.basename(fname)
        return c

    @staticmethod
    def from_qsim_file(fname: str) -> 'Circuit':
        """Produces a :class:`Circuit` based on a .qc description of a circuit.
        If a Toffoli gate with more than 2 controls is encountered, ancilla qubits are added.
        Currently up to 5 controls are supported."""
        from .qsimparser import parse_qsim
        with open(fname, 'r') as f:
            data = f.read()
        c = parse_qsim(data)
        c.name = os.path.basename(fname)
        return c

    @staticmethod
    def from_quipper(s: str) -> 'Circuit':
        """Produces a :class:`Circuit` based on a Quipper ASCII description of a circuit.
        Currently measurement instructions are not supported and are discarded."""
        from .quipperparser import parse_quipper_block
        text = s.strip()
        lines = text.splitlines()
        if text.find('Subroutine') == -1:
            c = parse_quipper_block(lines)
            return c
        else:
            raise TypeError("Subroutines are not supported")

    @staticmethod
    def from_quipper_file(fname: str) -> 'Circuit':
        """Produces a :class:`Circuit` based on a Quipper ASCII description of a circuit.
        Currently measurement instructions are not supported and are discarded."""
        from .quipperparser import parse_quipper_block, quipper_center_block
        try:
            with open(fname, 'r') as f:
                text = f.read().strip()
                lines = text.splitlines()
            if text.find('Subroutine') == -1:
                c = parse_quipper_block(lines)
                c.name = os.path.basename(fname)
                return c
            else:
                raise TypeError("Subroutines are not supported")
        except:
            return quipper_center_block(fname)

    @staticmethod
    def from_qasm(s: str) -> 'Circuit':
        """Produces a :class:`Circuit` based on a QASM input string.
        It ignores all the non-unitary instructions like measurements in the file.
        It currently doesn't support custom gates that have parameters."""
        from .qasmparser import QASMParser
        p = QASMParser()
        return p.parse(s)

    @staticmethod
    def from_qasm_file(fname: str) -> 'Circuit':
        """Produces a :class:`Circuit` based on a QASM description of a circuit.
        It ignores all the non-unitary instructions like measurements in the file.
        It currently doesn't support custom gates that have parameters."""
        from .qasmparser import QASMParser
        p = QASMParser()
        with open(fname, 'r') as f:
            c = p.parse(f.read())
        c.name = os.path.basename(fname)
        return c

    def to_quipper(self) -> str:
        """Produces a Quipper ASCII description of the circuit."""
        s = "Inputs: " + ", ".join("{!s}:Qbit".format(i) for i in range(self.qubits)) + "\n"
        for g in self.gates:
            s += g.to_quipper() + "\n"
        s += "Outputs: " + ", ".join("{!s}:Qbit".format(i) for i in range(self.qubits))
        return s

    def to_qasm(self, version: int = 2) -> str:
        """Produces a QASM description of the circuit."""
        assert version in [2, 3]
        if version == 3:
            s = """OPENQASM 3;\ninclude "stdgates.inc";\n"""
            s += "qubit[{!s}] q;\n".format(self.qubits)
        else:
            s = """OPENQASM 2.0;\ninclude "qelib1.inc";\n"""
            s += "qreg q[{!s}];\n".format(self.qubits)
        for g in self.gates:
            s += g.to_qasm() + "\n"
        return s

    def to_qc(self) -> str:
        """Produces a .qc description of the circuit."""
        s = ".v " + " ".join("q{:d}".format(i) for i in range(self.qubits))
        s += "\n\nBEGIN\n"
        c = self.split_phase_gates()
        for g in c.gates:
            s += g.to_qc() + "\n"
        s += "END\n"
        return s



    ### STAT FUNCTIONS


    def tcount(self) -> int:
        """Returns the amount of T-gates necessary to implement this circuit."""
        return sum(g.tcount() for g in self.gates)
        #return sum(1 for g in self.gates if isinstance(g, (ZPhase, XPhase, ParityPhase)) and g.phase.denominator >= 4)

    def twoqubitcount(self) -> int:
        """Returns the amount of 2-qubit gates necessary to implement this circuit."""
        c = self.to_basic_gates()
        return sum(1 for g in c.gates if g.name in ('CNOT','CZ'))

    def stats(self, depth: bool = False) -> str:
        """Returns statistics on the amount of gates in the circuit, separated into different classes
        (such as amount of T-gates, two-qubit gates, Hadamard gates)."""
        d = self.stats_dict(depth)
        s = """Circuit {} on {} qubits with {} gates.
        {} is the T-count
        {} Cliffords among which
        {} 2-qubit gates ({} CNOT, {} other) and
        {} Hadamard gates.""".format(d["name"], d["qubits"], d["gates"],
                d["tcount"], d["clifford"], d["twoqubit"], d["cnot"], d["twoqubit"] - d["cnot"], d["had"])
        if d["measurement"] > 0:
            s += "\nThere are {} measurement gates".format(d["measurement"])
        if d["other"] > 0:
            s += "\nThere are {} gates of a different type".format(d["other"])
        if depth:
            s += "\nThe circuit depth is {}".format(d["depth"])
            s += "\nThe circuit depth if no CZs are possible is {}".format(d["depth_cz"])
        return s

    def stats_dict(self, depth: bool = False) -> dict:
        """Returns a dictionary containing statistics on the amount of gates in the circuit,
        separated into different classes (such as amount of T-gates, two-qubit gates, Hadamard gates)."""
        total = 0
        tcount = 0
        twoqubit = 0
        hadamard = 0
        clifford = 0
        measurement = 0
        other = 0
        cnot = 0
        for g in self.gates:
            total += 1
            tcount += g.tcount()
            if isinstance(g, (ZPhase, XPhase)):
                if g.phase.denominator <= 2: clifford += 1
            elif isinstance(g, HAD):
                hadamard += 1
                clifford += 1
            elif isinstance(g, (CZ, XCX, CNOT)):
                twoqubit += 1
                clifford += 1
                if isinstance(g, CNOT): cnot += 1
            elif isinstance(g, Measurement):
                measurement += 1
            else:
                other += 1
        d : Dict[str, Union[str,int]] = dict()
        d["name"] = self.name
        d["qubits"] = self.qubits
        d["gates"] = total
        d["tcount"] = tcount
        d["clifford"] = clifford
        d["twoqubit"] = twoqubit
        d["cnot"] = cnot
        d["had"] = hadamard
        d["measurement"] = measurement
        d["other"] = other
        d["depth"] = 0
        d["depth_cz"] = 0
        if depth:
            c = Circuit(self.qubits)
            c.gates = [basic_gate for g in self.gates for basic_gate in g.to_basic_gates()]
            d["depth"] = c.depth()
            basic_no_cz = []
            for g in c.gates:
                if isinstance(g, CZ):
                    basic_no_cz.extend([HAD(g.target), CNOT(g.control, g.target), HAD(g.target)])
                else:
                    basic_no_cz.append(g)
            c.gates = basic_no_cz
            d["depth_cz"] = c.depth()
        return d

    def depth(self) -> int:
        min_depth = [0] * self.qubits
        for g in self.gates:
            if isinstance(g, (ZPhase, XPhase, HAD)):
                min_depth[g.target] += 1
            elif isinstance(g, (CZ, CNOT, SWAP)):
                gate_depth = max(min_depth[g.target], min_depth[g.control]) + 1
                min_depth[g.target] = gate_depth
                min_depth[g.control] = gate_depth
            elif isinstance(g, (CCZ, Tofolli)):
                gate_depth = max(min_depth[g.target], min_depth[g.ctrl1], min_depth[g.ctrl2]) + 1
                min_depth[g.target] = gate_depth
                min_depth[g.ctrl1] = gate_depth
                min_depth[g.ctrl2] = gate_depth
        return max(min_depth)


def determine_file_type(circuitfile: str) -> str:
        """Tries to figure out in which format the file is given (quipper, qasm or qc)"""
        fname = circuitfile
        ext = os.path.splitext(fname)[-1]
        if ext in ('.qc', '.tfc'):
            return "qc"
        if ext.find('qasm') != -1:
            return "qasm"
        if ext == '.qgraph':
            return "qgraph"
        if ext == '.qsim':
            return 'qsim'
        if ext.find('quip') != -1:
            return "quipper"
        f = open(fname, 'r')
        data = f.read(128)
        f.close()
        if data.startswith('Inputs:'):
            return "quipper"
        if data.find('.v') != -1 or data.find('.i') != -1 or data.find('.o') != -1:
            return "qc"
        if data.find('QASM') != -1:
            return "qasm"

        raise TypeError("Couldn't determine circuit format.")


def id(n: int) -> Circuit:
    return Circuit(n)



# PyZX - Python library for quantum circuit rewriting
#        and optimisation using the ZX-calculus
# Copyright (C) 2021 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



"""
This module contains two congruences (i.e., non-simplification rewrite rules) for exploring the space of equivalent ZX-diagrams. The two congruences defined here are based on the graph-theoretic notions of local complementation and pivoting. The methods lc_cong and pivot_cong take a ZX-diagram and subjects over which to apply the rewrite rule as parameters. The methods `apply_rand_lc` and `apply_rand_pivot` select these subjects probabilistically.
"""


import numpy as np
import itertools
from fractions import Fraction

from ..rewrite_rules.rules import apply_rule, lcomp

import sys
if __name__ == '__main__':
    sys.path.append('..')
from pyzx_param.utils import VertexType, EdgeType

### Utilities
def toggle_edge(g, v1, v2):
    """
    Utility function that toggles the connectivity between two spiders in a graph-like ZX-diagram.

    :param g: Graph where the operation is applied.
    :param v1: First vertex to toggle.
    :param v2: Second vertex to toggle
    """
    if g.connected(v1, v2):
        g.remove_edge(g.edge(v1, v2))
    else:
        g.add_edge((v1, v2), edgetype=EdgeType.HADAMARD)

def toggle_subset_connectivity(g, vs1, vs2):
    """
    Utility function that toggles the connectivity between two subsets of spiders in a graph-like ZX-diagram.

    :param g: Graph where the operation is applied.
    :param vs1: First subset of vertices to be toggled.
    :param vs2: Second subset of vertices to be toggled.

    """
    for v1 in vs1:
        for v2 in vs2:
            toggle_edge(g, v1, v2)

def uniform_weights(g, elts):
    """
    Assigns uniform weights to elements for selection.
    Used as weight function for random local complementation and random pivoting functions.

    :param g: Graph context for the operation.
    :param elts: Elements to assign weights to.
    :return: List of uniform weights.
    """
    return [1 / len(elts)] * len(elts)

def unfuse(g, v):
    """
    For a Z-spider with a phase and neighbors that are both BOUNDARY and Z, unfuse the phase
    in the form of a new spider that holds the connectivity to the boundaries. Note that v will
    maintain its connectivity to any Z spiders (just not boundaries).

    :param g: Graph where the operation is applied.
    :param v: Vertex to unfuse.
    :return: New vertex created during unfusion.
    """
    v_phase = g.phase(v)
    bs = [v for v in g.neighbors(v) if g.type(v) == VertexType.BOUNDARY]

    # we could enforce the following to ensure graph-likeness, but not necessary
    # assert(len(bs) < 2)

    new_v = g.add_vertex(
        ty=VertexType.Z,
        phase=v_phase)

    # unfuse the phase
    g.set_phase(v, phase=0)
    g.add_edge((v, new_v), edgetype=EdgeType.SIMPLE)

    # transfer over the boundary connectivity to the new spider
    for b in bs:
        e = g.edge(v, b)
        b_edge_type = g.edge_type(e)
        g.remove_edge(e)
        g.add_edge((new_v, b), edgetype=b_edge_type)

    # return the reference to the new vertex
    return new_v




### Local Complementation
def is_lc_vertex(g, v):
    """Checks if a spider in a ZX-diagram is a valid subject for local complementation.

    :param g: Graph where the check is performed.
    :param v: Vertex to check.
    :return: True if valid, False otherwise.
    """
    # don't want to apply LC to a single-degree spider because it will only result in a growing chain
    if g.vertex_degree(v) < 2 or g.type(v) != VertexType.Z:
        return False

    # TODO: no I/O vertices for now. note that they are included in degree count
    for n in g.neighbors(v):
        if g.type(n) == VertexType.BOUNDARY:
            return False

    return True


# Assumes that v is a Z-spider (green)
def lc_cong(g, v):
    """Applies local complementation at a provided spider in a ZX-diagram. Assumes v is a Z spider.

    :param g: Graph where the operation is applied.
    :param v: Vertex to apply local complementation.
    """
    # FIXME: not gracefully handling if on boundary. If on boundary, sohuld just add on same qubit rather than add a gadget
    ns = [n for n in g.neighbors(v) if g.type(n) == VertexType.Z]

    # complement edges between neighbors
    # TODO: use add_edge_table
    for n1, n2 in itertools.combinations(ns, 2):
        toggle_edge(g, n1, n2)

    # add pi/2 to all neighbors
    for n in ns:
        g.add_to_phase(n, Fraction(1, 2))

    # add gadget (and transfer over the phase)
    new_v = g.add_vertex(
        ty=VertexType.Z,
        phase=g.phase(v) + Fraction(1, 2),
        qubit=g.qubit(v)-1,
        row=g.row(v)
    )
    g.set_phase(v, phase=Fraction(1, 2))
    g.add_edge((v, new_v), edgetype=EdgeType.HADAMARD)


def lc_cong2(g, v):
    """Applies local complementation at a provided spider in a ZX-diagram.

    :param g: Graph where the operation is applied.
    :param v: Vertex to apply local complementation.
    """
    p = g.phase(v)
    if p != Fraction(1,2) and p != Fraction(-1,2):
        v1 = g.add_vertex(
                ty=VertexType.Z,
                phase=g.phase(v) + Fraction(1, 2),
                qubit=g.qubit(v)-2,
                row=g.row(v))
        v2 = g.add_vertex(
                ty=VertexType.Z,
                qubit=g.qubit(v)-1,
                row=g.row(v))
        g.add_edge((v1, v2), edgetype=EdgeType.HADAMARD)
        g.add_edge((v2, v), edgetype=EdgeType.HADAMARD)
        g.set_phase(v, Fraction(-1,2))
    # apply_rule(g, lcomp, [[v, list(g.neighbors(v))]])



def apply_rand_lc(g, weight_func=uniform_weights):
    """Applies local complementation to randomly selected spider.

    :param g: Graph where the operation is applied.
    :param weight_func: Function to determine weights for selection.    
    """

    lc_vs = [v for v in g.vertices() if is_lc_vertex(g, v)]
    weights = weight_func(g, lc_vs)
    lc_v = np.random.choice(lc_vs, 1, p=weights)[0]
    lc_cong2(g, lc_v)



### Pivoting


# TODO: May want to add some additional cases when we know it's not useful (as in the LC case)
def is_pivot_edge(g, e):
    """
    Checks if a given edge in a ZX-diagram is a suitable candidate for pivoting
    
    :param g: Graph where the check is performed.
    :param e: Edge to check.
    :return: True if suitable, False otherwise.
    """

    v1, v2 = g.edge_st(e)
    return g.type(v1) == VertexType.Z and g.type(v2) == VertexType.Z

def pivot_cong(g, v1, v2):
    """
    Applies pivoting to two connected spiders

    :param g: Graph where the operation is applied.
    :param v1: First vertex.
    :param v2: Second vertex.
    """

    # get the three subsets
    nhd1 = list(g.neighbors(v1))
    nhd2 = list(g.neighbors(v2))
    assert(all([g.type(v) in [VertexType.Z, VertexType.BOUNDARY] for v in nhd1]))
    assert(all([g.type(v) in [VertexType.Z, VertexType.BOUNDARY] for v in nhd2]))
    vs1 = [v for v in nhd1 if g.type(v) == VertexType.Z]
    bs1 = [v for v in nhd1 if g.type(v) == VertexType.BOUNDARY]
    vs2 = [v for v in nhd2 if g.type(v) == VertexType.Z]
    bs2 = [v for v in nhd2 if g.type(v) == VertexType.BOUNDARY]

    shared_ns = list(set(vs1) & set(vs2))
    nhd1_only = list(set(vs1) - set(vs2 + [v2]))
    nhd2_only = list(set(vs2) - set(vs1 + [v1]))

    # toggle connectivity between the three subsets
    toggle_subset_connectivity(g, shared_ns, nhd1_only)
    toggle_subset_connectivity(g, shared_ns, nhd2_only)
    toggle_subset_connectivity(g, nhd1_only, nhd2_only)

    # add pi to all the shared neighbors
    for n in shared_ns:
        g.add_to_phase(n, 1)

    # for v1 and v2, pull out phases along hadamards and swap (while maintaining any I/O)
    new_v1 = unfuse(g, v1)
    new_v2 = unfuse(g, v2)

    g.remove_edge(g.edge(v1, new_v1))
    g.remove_edge(g.edge(v2, new_v2))
    g.add_edge((v1, new_v2), edgetype=EdgeType.HADAMARD)
    g.add_edge((v2, new_v1), edgetype=EdgeType.HADAMARD)




def apply_rand_pivot(g, weight_func=uniform_weights):
    """
    Applies pivoting to a randomly selected pair of connected spiders.

    :param g: Graph where the operation is applied.
    :param weight_func: Function to determine weights for selection. Default is uniform weights.
    """

    # assumes len(candidates) != 0
    candidates = [e for e in g.edges() if is_pivot_edge(g, e)]
    weights = weight_func(g, candidates)
    e_idx = np.random.choice(len(candidates), 1, p=weights)[0]
    e = candidates[e_idx]
    v1, v2 = g.edge_st(e)
    pivot_cong(g, v1, v2)
